from dataclasses import dataclass
from enum import auto
from typing import Generic, Sequence, TypeVar

import pytest
from connexion import FlaskApp
from flask_login import UserMixin
from injector import Module
from Ligare.platform.dependency_injection import UserLoaderModule
from Ligare.platform.feature_flag import FeatureFlag
from Ligare.platform.feature_flag.feature_flag_router import FeatureFlagRouter
from Ligare.platform.identity.user_loader import Role as LoaderRole
from Ligare.programming.config import AbstractConfig
from Ligare.web.application import CreateAppResult, OpenAPIAppResult
from Ligare.web.config import Config
from Ligare.web.middleware.feature_flags import (
    CachingFeatureFlagRouterModule,
    FeatureFlagConfig,
    FeatureFlagMiddlewareModule,
)
from Ligare.web.testing.create_app import (
    CreateOpenAPIApp,
    OpenAPIClientInjectorConfigurable,
    OpenAPIMockController,
)
from mock import MagicMock
from pytest_mock import MockerFixture
from typing_extensions import override


@dataclass
class UserId:
    user_id: int
    username: str


class Role(LoaderRole):
    User = auto()
    Administrator = auto()
    Operator = auto()

    @staticmethod
    def items():
        return Role.__members__.items()


TRole = TypeVar("TRole", bound=Role, covariant=True)


class User(UserMixin, Generic[TRole]):
    """
    Represents the user object stored in a session.
    """

    id: UserId
    roles: Sequence[TRole]

    @override
    def get_id(self):
        """
        Override the UserMixin.get_id so the username is returned instead of `id` (the dataclass)
        when `flask_login.login_user` calls this method to assign the
        session `_user_id` key.
        """
        return str(self.id.username)

    @override
    def __init__(self, id: UserId, roles: Sequence[TRole] | None = None):
        """
        Create a new user with the given user name or id, and a list of roles.
        If roles are not given, an empty list is assigned by default.
        """
        super().__init__()

        if roles is None:
            roles = []

        self.id = id
        self.roles = roles


class TestFeatureFlagsMiddleware(CreateOpenAPIApp):
    def _user_session_app_init_hook(
        self,
        application_configs: list[type[AbstractConfig]],
        application_modules: list[Module | type[Module]],
    ):
        application_modules.append(
            UserLoaderModule(
                loader=User,  # pyright: ignore[reportArgumentType]
                roles=Role,
                user_table=MagicMock(),  # pyright: ignore[reportArgumentType]
                role_table=MagicMock(),  # pyright: ignore[reportArgumentType]
                bases=[],
            )
        )
        application_modules.append(CachingFeatureFlagRouterModule)
        application_modules.append(FeatureFlagMiddlewareModule())

    def test__FeatureFlagMiddleware__feature_flag_api_GET_requires_user_session_when_flask_login_is_configured(
        self,
        openapi_config: Config,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_mock_controller: OpenAPIMockController,
        mocker: MockerFixture,
    ):
        def app_init_hook(
            application_configs: list[type[AbstractConfig]],
            application_modules: list[Module | type[Module]],
        ):
            application_modules.append(CachingFeatureFlagRouterModule)
            application_modules.append(FeatureFlagMiddlewareModule())

        openapi_mock_controller.begin()
        app = next(
            openapi_client_configurable(
                openapi_config,
                app_init_hook=app_init_hook,
            )
        )

        response = app.client.get("/platform/feature_flag")

        assert response.status_code == 401

    @pytest.mark.parametrize(
        "flask_login_is_configured,user_has_session",
        [[True, True], [False, True], [False, False]],
    )
    def test__FeatureFlagMiddleware__feature_flag_api_GET_gets_feature_flags(
        self,
        flask_login_is_configured: bool,
        user_has_session: bool,
        openapi_config: Config,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_mock_controller: OpenAPIMockController,
        mocker: MockerFixture,
    ):
        def app_init_hook(
            application_configs: list[type[AbstractConfig]],
            application_modules: list[Module | type[Module]],
        ):
            if not flask_login_is_configured:
                application_modules.clear()
            application_modules.append(CachingFeatureFlagRouterModule)
            application_modules.append(FeatureFlagMiddlewareModule())

        def client_init_hook(app: CreateAppResult[FlaskApp]):
            caching_feature_flag_router = app.app_injector.flask_injector.injector.get(
                FeatureFlagRouter[FeatureFlag]
            )
            _ = caching_feature_flag_router.set_feature_is_enabled("foo_feature", True)

        openapi_mock_controller.begin()
        app = next(
            openapi_client_configurable(
                openapi_config,
                client_init_hook=client_init_hook,
                app_init_hook=app_init_hook,
            )
        )

        if user_has_session:
            with self.get_authenticated_request_context(
                app,
                User,  # pyright: ignore[reportArgumentType]
                mocker,
            ):
                response = app.client.get("/platform/feature_flag")
        else:
            response = app.client.get("/platform/feature_flag")

        assert response.status_code == 200
        response_json = response.json()
        assert (data := response_json.get("data", None)) is not None
        assert len(data) == 1
        assert (name := data[0].get("name", None)) is not None
        assert (enabled := data[0].get("enabled", None)) is not None
        assert name == "foo_feature"
        assert enabled == True

    @pytest.mark.parametrize("has_role", [True, False])
    def test__FeatureFlagMiddleware__feature_flag_api_GET_requires_specified_role_when_flask_login_is_configured(
        self,
        has_role: bool,
        openapi_config: Config,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_mock_controller: OpenAPIMockController,
        mocker: MockerFixture,
    ):
        get_feature_flag_mock = mocker.patch(
            "Ligare.web.middleware.feature_flags.CachingFeatureFlagRouter.get_feature_flags",
            return_value=[],
        )

        def client_init_hook(app: OpenAPIAppResult):
            feature_flag_config = FeatureFlagConfig(
                access_role_name="Operator",
                api_base_url="/platform",  # the default
            )
            app.app_injector.flask_injector.injector.binder.bind(
                FeatureFlagConfig, to=feature_flag_config
            )

        openapi_mock_controller.begin()
        app = next(
            openapi_client_configurable(
                openapi_config, client_init_hook, self._user_session_app_init_hook
            )
        )

        with self.get_authenticated_request_context(
            app,
            User,  # pyright: ignore[reportArgumentType]
            mocker,
            [Role.Operator] if has_role else [],
        ):
            response = app.client.get("/platform/feature_flag")

        if has_role:
            assert response.status_code == 404
            get_feature_flag_mock.assert_called_once()
        else:
            assert response.status_code == 401
            get_feature_flag_mock.assert_not_called()

    def test__FeatureFlagMiddleware__feature_flag_api_GET_returns_no_feature_flags_when_none_exist(
        self,
        openapi_config: Config,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_mock_controller: OpenAPIMockController,
        mocker: MockerFixture,
    ):
        openapi_mock_controller.begin()
        app = next(
            openapi_client_configurable(
                openapi_config, app_init_hook=self._user_session_app_init_hook
            )
        )

        with self.get_authenticated_request_context(
            app,
            User,  # pyright: ignore[reportArgumentType]
            mocker,
        ):
            response = app.client.get("/platform/feature_flag")

        assert response.status_code == 404
        response_json = response.json()
        assert (problems := response_json.get("problems", None)) is not None
        assert len(problems) == 1
        assert (title := problems[0].get("title", None)) is not None
        assert title == "No feature flags found"

    def test__FeatureFlagMiddleware__feature_flag_api_GET_returns_feature_flags_when_they_exist(
        self,
        openapi_config: Config,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_mock_controller: OpenAPIMockController,
        mocker: MockerFixture,
    ):
        def client_init_hook(app: CreateAppResult[FlaskApp]):
            caching_feature_flag_router = app.app_injector.flask_injector.injector.get(
                FeatureFlagRouter[FeatureFlag]
            )
            _ = caching_feature_flag_router.set_feature_is_enabled("foo_feature", True)

        openapi_mock_controller.begin()
        app = next(
            openapi_client_configurable(
                openapi_config,
                client_init_hook,
                self._user_session_app_init_hook,
            )
        )

        with self.get_authenticated_request_context(
            app,
            User,  # pyright: ignore[reportArgumentType]
            mocker,
        ):
            response = app.client.get("/platform/feature_flag")

        assert response.status_code == 200
        response_json = response.json()
        assert (data := response_json.get("data", None)) is not None
        assert len(data) == 1
        assert data[0].get("enabled", None) is True
        assert data[0].get("name", None) == "foo_feature"

    @pytest.mark.parametrize(
        "query_flags", ["bar_feature", ["foo_feature", "baz_feature"]]
    )
    def test__FeatureFlagMiddleware__feature_flag_api_GET_returns_specific_feature_flags_when_they_exist(
        self,
        query_flags: str | list[str],
        openapi_config: Config,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_mock_controller: OpenAPIMockController,
        mocker: MockerFixture,
    ):
        def client_init_hook(app: CreateAppResult[FlaskApp]):
            caching_feature_flag_router = app.app_injector.flask_injector.injector.get(
                FeatureFlagRouter[FeatureFlag]
            )
            _ = caching_feature_flag_router.set_feature_is_enabled("foo_feature", True)
            _ = caching_feature_flag_router.set_feature_is_enabled("bar_feature", False)
            _ = caching_feature_flag_router.set_feature_is_enabled("baz_feature", True)

        openapi_mock_controller.begin()
        app = next(
            openapi_client_configurable(
                openapi_config,
                client_init_hook,
                self._user_session_app_init_hook,
            )
        )

        with self.get_authenticated_request_context(
            app,
            User,  # pyright: ignore[reportArgumentType]
            mocker,
        ):
            response = app.client.get(
                "/platform/feature_flag", params={"name": query_flags}
            )

        assert response.status_code == 200
        response_json = response.json()
        assert (data := response_json.get("data", None)) is not None
        if isinstance(query_flags, str):
            assert len(data) == 1
            assert data[0].get("enabled", None) is False
            assert data[0].get("name", None) == query_flags
        else:
            assert len(data) == len(query_flags)
            for i, flag in enumerate(query_flags):
                assert data[i].get("enabled", None) is True
                assert data[i].get("name", None) == flag

    @pytest.mark.parametrize(
        "flask_login_is_configured,user_has_session,error_code",
        [[True, False, 401], [False, True, 405], [False, False, 405]],
    )
    def test__FeatureFlagMiddleware__feature_flag_api_PATCH_requires_user_session_and_flask_login(
        self,
        flask_login_is_configured: bool,
        user_has_session: bool,
        error_code: int,
        openapi_config: Config,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_mock_controller: OpenAPIMockController,
        mocker: MockerFixture,
    ):
        set_feature_flag_mock = mocker.patch(
            "Ligare.web.middleware.feature_flags.CachingFeatureFlagRouter.set_feature_is_enabled",
            return_value=[],
        )

        def app_init_hook(
            application_configs: list[type[AbstractConfig]],
            application_modules: list[Module | type[Module]],
        ):
            if not flask_login_is_configured:
                application_modules.clear()
            application_modules.append(CachingFeatureFlagRouterModule)
            application_modules.append(FeatureFlagMiddlewareModule())

        openapi_mock_controller.begin()
        app = next(
            openapi_client_configurable(
                openapi_config,
                app_init_hook=app_init_hook,
            )
        )

        if user_has_session:
            with self.get_authenticated_request_context(
                app,
                User,  # pyright: ignore[reportArgumentType]
                mocker,
            ):
                response = app.client.patch(
                    "/platform/feature_flag",
                    json=[{"name": "foo_feature", "enabled": False}],
                )
        else:
            response = app.client.patch(
                "/platform/feature_flag",
                json=[{"name": "foo_feature", "enabled": False}],
            )

        assert response.status_code == error_code
        set_feature_flag_mock.assert_not_called()

    def test__FeatureFlagMiddleware__feature_flag_api_PATCH_modifies_feature_flag_when_user_has_session_and_flask_login_is_configured(
        self,
        openapi_config: Config,
        openapi_client_configurable: OpenAPIClientInjectorConfigurable,
        openapi_mock_controller: OpenAPIMockController,
        mocker: MockerFixture,
    ):
        def client_init_hook(app: CreateAppResult[FlaskApp]):
            caching_feature_flag_router = app.app_injector.flask_injector.injector.get(
                FeatureFlagRouter[FeatureFlag]
            )
            _ = caching_feature_flag_router.set_feature_is_enabled("foo_feature", True)

        openapi_mock_controller.begin()
        app = next(
            openapi_client_configurable(
                openapi_config,
                client_init_hook,
                self._user_session_app_init_hook,
            )
        )

        with self.get_authenticated_request_context(
            app,
            User,  # pyright: ignore[reportArgumentType]
            mocker,
        ):
            response = app.client.patch(
                "/platform/feature_flag",
                json=[{"name": "foo_feature", "enabled": False}],
            )

        assert response.status_code == 200
        response_json = response.json()
        assert (data := response_json.get("data", None)) is not None
        assert len(data) == 1
        assert data[0].get("name", None) == "foo_feature"
        assert data[0].get("new_value", None) == False
        assert data[0].get("old_value", None) == True



"""
This type stub file was generated by pyright.
"""

from functools import _PWrapped, _Wrapped
from typing import Any, Callable, TypeVar

from werkzeug.local import LocalProxy

T = TypeVar("T")

current_user: LocalProxy[T] = ...

def encode_cookie(payload, key=...): # -> str:
    """
    This will encode a ``str`` value into a cookie, and sign that cookie
    with the app's secret key.

    :param payload: The value to encode, as `str`.
    :type payload: str

    :param key: The key to use when creating the cookie digest. If not
                specified, the SECRET_KEY value from app config will be used.
    :type key: str
    """
    ...

def decode_cookie(cookie, key=...): # -> None:
    """
    This decodes a cookie given by `encode_cookie`. If verification of the
    cookie fails, ``None`` will be implicitly returned.

    :param cookie: An encoded cookie.
    :type cookie: str

    :param key: The key to use when creating the cookie digest. If not
                specified, the SECRET_KEY value from app config will be used.
    :type key: str
    """
    ...

def make_next_param(login_url, current_url): # -> str:
    """
    Reduces the scheme and host from a given URL so it can be passed to
    the given `login` URL more efficiently.

    :param login_url: The login URL being redirected to.
    :type login_url: str
    :param current_url: The URL to reduce.
    :type current_url: str
    """
    ...

def expand_login_view(login_view): # -> str:
    """
    Returns the url for the login view, expanding the view name to a url if
    needed.

    :param login_view: The name of the login view or a URL for the login view.
    :type login_view: str
    """
    ...

def login_url(login_view, next_url=..., next_field=...): # -> str:
    """
    Creates a URL for redirecting to a login page. If only `login_view` is
    provided, this will just return the URL for it. If `next_url` is provided,
    however, this will append a ``next=URL`` parameter to the query string
    so that the login view can redirect back to that URL. Flask-Login's default
    unauthorized handler uses this function when redirecting to your login url.
    To force the host name used, set `FORCE_HOST_FOR_REDIRECTS` to a host. This
    prevents from redirecting to external sites if request headers Host or
    X-Forwarded-For are present.

    :param login_view: The name of the login view. (Alternately, the actual
                       URL to the login view.)
    :type login_view: str
    :param next_url: The URL to give the login view for redirection.
    :type next_url: str
    :param next_field: What field to store the next URL in. (It defaults to
                       ``next``.)
    :type next_field: str
    """
    ...

def login_fresh(): # -> bool:
    """
    This returns ``True`` if the current login is fresh.
    """
    ...

def login_remembered(): # -> bool:
    """
    This returns ``True`` if the current login is remembered across sessions.
    """
    ...

def login_user(user, remember=..., duration=..., force=..., fresh=...): # -> bool:
    """
    Logs a user in. You should pass the actual user object to this. If the
    user's `is_active` property is ``False``, they will not be logged in
    unless `force` is ``True``.

    This will return ``True`` if the log in attempt succeeds, and ``False`` if
    it fails (i.e. because the user is inactive).

    :param user: The user object to log in.
    :type user: object
    :param remember: Whether to remember the user after their session expires.
        Defaults to ``False``.
    :type remember: bool
    :param duration: The amount of time before the remember cookie expires. If
        ``None`` the value set in the settings is used. Defaults to ``None``.
    :type duration: :class:`datetime.timedelta`
    :param force: If the user is inactive, setting this to ``True`` will log
        them in regardless. Defaults to ``False``.
    :type force: bool
    :param fresh: setting this to ``False`` will log in the user with a session
        marked as not "fresh". Defaults to ``True``.
    :type fresh: bool
    """
    ...

def logout_user(): # -> Literal[True]:
    """
    Logs a user out. (You do not need to pass the actual user.) This will
    also clean up the remember me cookie if it exists.
    """
    ...

def confirm_login(): # -> None:
    """
    This sets the current session as fresh. Sessions become stale when they
    are reloaded from a cookie.
    """
    ...

def login_required(
    func: Callable[..., Any]
) -> _Wrapped[_PWrapped, Any, _PWrapped, Any]:
    """
    If you decorate a view with this, it will ensure that the current user is
    logged in and authenticated before calling the actual view. (If they are
    not, it calls the :attr:`LoginManager.unauthorized` callback.) For
    example::

        @app.route('/post')
        @login_required
        def post():
            pass

    If there are only certain times you need to require that your user is
    logged in, you can do so with::

        if not current_user.is_authenticated:
            return current_app.login_manager.unauthorized()

    ...which is essentially the code that this function adds to your views.

    It can be convenient to globally turn off authentication when unit testing.
    To enable this, if the application configuration variable `LOGIN_DISABLED`
    is set to `True`, this decorator will be ignored.

    .. Note ::

        Per `W3 guidelines for CORS preflight requests
        <http://www.w3.org/TR/cors/#cross-origin-request-with-preflight-0>`_,
        HTTP ``OPTIONS`` requests are exempt from login checks.

    :param func: The view function to decorate.
    :type func: function
    """
    ...

def fresh_login_required(func): # -> _Wrapped[Callable[..., Any], Any, Callable[..., Any], Any]:
    """
    If you decorate a view with this, it will ensure that the current user's
    login is fresh - i.e. their session was not restored from a 'remember me'
    cookie. Sensitive operations, like changing a password or e-mail, should
    be protected with this, to impede the efforts of cookie thieves.

    If the user is not authenticated, :meth:`LoginManager.unauthorized` is
    called as normal. If they are authenticated, but their session is not
    fresh, it will call :meth:`LoginManager.needs_refresh` instead. (In that
    case, you will need to provide a :attr:`LoginManager.refresh_view`.)

    Behaves identically to the :func:`login_required` decorator with respect
    to configuration variables.

    .. Note ::

        Per `W3 guidelines for CORS preflight requests
        <http://www.w3.org/TR/cors/#cross-origin-request-with-preflight-0>`_,
        HTTP ``OPTIONS`` requests are exempt from login checks.

    :param func: The view function to decorate.
    :type func: function
    """
    ...

def set_login_view(login_view, blueprint=...): # -> None:
    """
    Sets the login view for the app or blueprint. If a blueprint is passed,
    the login view is set for this blueprint on ``blueprint_login_views``.

    :param login_view: The user object to log in.
    :type login_view: str
    :param blueprint: The blueprint which this login view should be set on.
        Defaults to ``None``.
    :type blueprint: object
    """
    ...




"""
This type stub file was generated by pyright.
"""

from connexion.apps.asynchronous import AsyncApi, AsyncApp
from connexion.apps.flask import FlaskApi, FlaskApp
from connexion.context import request
from connexion.middleware import ConnexionMiddleware

from .apps import AbstractApp
from .apps.asynchronous import AsyncApp
from .datastructures import NoContent
from .exceptions import ProblemException
from .problem import problem
from .resolver import Resolution, Resolver, RestyResolver
from .utils import not_installed_error

"""
Connexion is a framework that automagically handles HTTP requests based on OpenAPI Specification
(formerly known as Swagger Spec) of your API described in YAML format. Connexion allows you to
write an OpenAPI specification, then maps the endpoints to your Python functions; this makes it
unique, as many tools generate the specification based on your Python code. You can describe your
REST API in as much detail as you want; then Connexion guarantees that it will work as you
specified.
"""
App = ...
Api = ...

__all__ = (
    "AsyncApi",
    "AsyncApp",
    "FlaskApi",
    "FlaskApp",
    "request",
    "ConnexionMiddleware",
    "AbstractApp",
    "AsyncApp",
    "NoContent",
    "ProblemException",
    "problem",
    "Resolution",
    "Resolver",
    "RestyResolver",
    "not_installed_error",
    "App",
    "Api",
)



"""
This type stub file was generated by pyright.
"""

import typing as t
from starlette.types import ASGIApp, Receive, Scope, Send
from connexion.exceptions import ProblemException
from connexion.middleware.abstract import RoutedAPI, RoutedMiddleware
from connexion.operations import AbstractOperation
from connexion.security import SecurityHandlerFactory
from connexion.spec import Specification

logger = ...
class SecurityOperation:
    def __init__(self, next_app: ASGIApp, *, security_handler_factory: SecurityHandlerFactory, security: list, security_schemes: dict) -> None:
        ...
    
    @classmethod
    def from_operation(cls, operation: t.Union[AbstractOperation, Specification], *, next_app: ASGIApp, security_handler_factory: SecurityHandlerFactory) -> SecurityOperation:
        """Create a SecurityOperation from an Operation of Specification instance

        :param operation: The operation can be both an Operation or Specification instance here
            since security is defined at both levels in the OpenAPI spec. Creating a
            SecurityOperation based on a Specification can be used to create a SecurityOperation
            for routes not explicitly defined in the specification.
        :param next_app: The next ASGI app to call.
        :param security_handler_factory: The factory to be used to generate security handlers for
            the different security schemes.
        """
        ...
    
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        ...
    


class SecurityAPI(RoutedAPI[SecurityOperation]):
    def __init__(self, *args, auth_all_paths: bool = ..., security_map: dict = ..., **kwargs) -> None:
        ...
    
    def add_auth_on_not_found(self) -> None:
        """Register a default SecurityOperation for routes that are not found."""
        ...
    
    def make_operation(self, operation: t.Union[AbstractOperation, Specification]) -> SecurityOperation:
        """Create a SecurityOperation from an Operation of Specification instance

        :param operation: The operation can be both an Operation or Specification instance here
            since security is defined at both levels in the OpenAPI spec. Creating a
            SecurityOperation based on a Specification can be used to create a SecurityOperation
            for routes not explicitly defined in the specification.
        """
        ...
    


class SecurityMiddleware(RoutedMiddleware[SecurityAPI]):
    """Middleware to check if operation is accessible on scope."""
    api_cls = SecurityAPI


class MissingSecurityOperation(ProblemException):
    ...





"""
This type stub file was generated by pyright.
"""

import typing as t
from starlette.middleware.errors import ServerErrorMiddleware as StarletteServerErrorMiddleware
from starlette.types import ASGIApp
from connexion.lifecycle import ConnexionRequest, ConnexionResponse
from connexion.middleware.exceptions import connexion_wrapper
from connexion.types import MaybeAwaitable

logger = ...
class ServerErrorMiddleware(StarletteServerErrorMiddleware):
    """Subclass of starlette ServerErrorMiddleware to change handling of Unhandled Server
    exceptions to existing connexion behavior."""
    def __init__(self, next_app: ASGIApp, handler: t.Optional[t.Callable[[ConnexionRequest, Exception], MaybeAwaitable[ConnexionResponse]]] = ...) -> None:
        ...
    
    @staticmethod
    @connexion_wrapper
    def error_response(_request: ConnexionRequest, exc: Exception) -> ConnexionResponse:
        """Default handler for any unhandled Exception"""
        ...
    





"""
This type stub file was generated by pyright.
"""

import typing as t
from starlette.types import ASGIApp, Receive, Scope, Send
from connexion.middleware.abstract import RoutedAPI, RoutedMiddleware
from connexion.operations import AbstractOperation

"""
Validation Middleware.
"""
logger = ...
class RequestValidationOperation:
    def __init__(self, next_app: ASGIApp, *, operation: AbstractOperation, strict_validation: bool = ..., validator_map: t.Optional[dict] = ...) -> None:
        ...
    
    def extract_content_type(self, headers: t.List[t.Tuple[bytes, bytes]]) -> t.Tuple[str, str]:
        """Extract the mime type and encoding from the content type headers.

        :param headers: Headers from ASGI scope

        :return: A tuple of mime type, encoding
        """
        ...
    
    def validate_mime_type(self, mime_type: str) -> None:
        """Validate the mime type against the spec if it defines which mime types are accepted.

        :param mime_type: mime type from content type header
        """
        ...
    
    @property
    def security_query_params(self) -> t.List[str]:
        """Get the names of query parameters that are used for security."""
        ...
    
    async def __call__(self, scope: Scope, receive: Receive, send: Send): # -> None:
        ...
    


class RequestValidationAPI(RoutedAPI[RequestValidationOperation]):
    """Validation API."""
    def __init__(self, *args, strict_validation=..., validator_map=..., uri_parser_class=..., **kwargs) -> None:
        ...
    
    def make_operation(self, operation: AbstractOperation) -> RequestValidationOperation:
        ...
    


class RequestValidationMiddleware(RoutedMiddleware[RequestValidationAPI]):
    """Middleware for validating requests according to the API contract."""
    api_cls = RequestValidationAPI


class MissingValidationOperation(Exception):
    """Missing validation operation"""
    ...





import logging
import re
import sys
import types
from dataclasses import asdict, dataclass, field
from importlib.machinery import SourceFileLoader
from importlib.util import module_from_spec, spec_from_file_location
from os import sep as path_separator
from pathlib import Path
from typing import Any, Literal, final

from jinja2 import BaseLoader, Environment, PackageLoader, Template
from Ligare.programming.collections.dict import merge

# fmt: off
from pkg_resources import IResourceProvider, ResourceManager, get_provider
from typing_extensions import override

# fmt: on


@dataclass
class Operation:
    url_path_name: str
    module_name: str
    raw_name: str

    def __init__(  # pyright: ignore[reportMissingSuperCall]
        self, name: str, **kwargs: Any | None
    ) -> None:
        name_lower = name.lower()
        self.raw_name = name
        self.url_path_name = name_lower
        self.module_name = re.sub(r"[^\w_]", "_", name_lower)

    @override
    def __str__(self) -> str:
        return self.url_path_name

    @override
    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, Operation) and (
            self.url_path_name == __value.url_path_name
            or self.module_name == __value.module_name
        )


@dataclass
class ScaffoldEndpoint:
    operation: Operation
    # This is Flask's default hostname.
    hostname: str = "http://localhost:5000"


@dataclass
class ScaffoldModule:
    module_name: str


@dataclass
class ScaffoldConfig:
    output_directory: str
    application: Operation
    template_type: Literal["basic", "openapi"] | None = None
    modules: list[ScaffoldModule] | None = None
    module: dict[str, Any] = field(default_factory=dict)
    endpoints: list[ScaffoldEndpoint] | None = None
    mode: str = "create"


@final
class Scaffolder:
    # These are used to aid in discovering files that are
    # part of the Ligare.web module. Relative path lookups
    # do not work, so they are done using _provider.
    _manager: Any
    _provider: IResourceProvider

    def __init__(self, config: ScaffoldConfig, log: logging.Logger) -> None:
        self._config = config
        self._config_dict = asdict(config)
        self._log = log
        self._checked_directories: set[Path] = set()
        # BaseLoader is used for rendering strings only. It does not need additional functionality.
        self._base_env = Environment(
            loader=BaseLoader  # pyright: ignore[reportArgumentType]
        )
        self._manager = ResourceManager()
        self._provider = get_provider("Ligare.web")

    def _create_directory(self, directory: Path, overwrite_existing_files: bool = True):
        """Create the directories that the rendered templates will be stored in."""
        if directory in self._checked_directories:
            return

        self._checked_directories.add(directory)
        if overwrite_existing_files and directory.exists():
            self._log.warning(
                f"Directory `{directory}` exists. Files may be overwritten."
            )
        else:
            self._log.debug(f"Creating directory `{directory}`.")
            # the "overwrite" check only applies in this method
            # to log a message. It is okay if a directory already exists.
            directory.mkdir(parents=True, exist_ok=True)

    def _render_template_string(
        self, template_string: str, template_config: dict[str, Any] | None = None
    ):
        """Render a string that may contain jinja2 directives."""
        if template_config is None:
            template_config = self._config_dict

        self._log.debug(
            f"Rendering template string `{template_string}` with config `{template_config}`."
        )

        rendered_string = self._base_env.from_string(template_string).render(
            **template_config
        )

        self._log.debug(
            f"Rendered template string `{template_string}` to `{rendered_string}`."
        )

        return rendered_string

    def _render_template(
        self,
        template_name: str,
        template_environment: Environment,
        template_directory_prefix: str = "",
        template_config: dict[str, Any] | None = None,
        template: Template | None = None,
        overwrite_existing_files: bool = True,
        write_rendered_template: bool = True,
    ):
        """
        Render a template that can either be found by name in
        the provided `template_environment`, or render the provided `template`.

        :param template_name: The name of the template that might be discoverable in the `template_environment`.
        :param template_environment: The jinja2 template environment used for rendering the template.
        :param template_directory_prefix: If not provided, templates will be stored relative to their discovered root in the template environment.
            For example, if the discovered template relative to the template environment exists at `./__init__.py`, the rendered template will be
            stored at `<output directory>/__init__.py`. If the template environment root points to a deeper directory in the template hierarchy,
            for example, in `{{application.module_name}}/modules/database/`, and the rendered template is `./__init__.py`, this default behavior
            is undesired. To resolve this, `template_directory_prefix` can be set in order to cause the rendered template to be stored under a
            different directory. For example, with the `template_directory_prefix` of `optional/{{application.module_name}}/modules/database/`, if the
            file `./__init__.py` is discovered at the template environment root,
            the file will instead be stored at `<output directory>/{{application.module_name}}/modules/database/__init__.py`.
        :param template_config: A dictionary of values that the template can reference.
        :param template: If provided, this template will be rendered instead of trying to discover the template with the name `template_name`
            in the template environment.
        :param overwrite_existing_files: If True, any existing files will be overwritten. If False, the template will not be rendered, and
            the file at the output location will not be overwritten.
        :param write_rendered_template: If True, the template will be written to the filesystem. If False, the template will be rendered, but
            not output anywhere.
        """
        if template_config is None:
            template_config = self._config_dict

        rendered_template_path = Path(
            # This causes paths with jinja2 directives to be rendered.
            # For example, if `self._config_dict['application'].module_name` is `foo`,
            # the path `base/{{application.module_name}}` is rendered as `base/foo`.
            self._render_template_string(template_name).replace(".j2", "")
        )
        # strip any leading / or c:/ etc. which should not occur here
        if rendered_template_path.is_absolute():
            rendered_template_path = path_separator.join(
                rendered_template_path.parts[1:]
            )

        template_output_path = Path(
            self._config.output_directory,
            template_directory_prefix,
            rendered_template_path,
        )

        if overwrite_existing_files == False and template_output_path.exists():
            self._log.warning(
                f"File `{template_output_path}` exists, but refusing to overwrite."
            )
            return

        self._create_directory(
            template_output_path.parent,
            overwrite_existing_files=overwrite_existing_files,
        )

        self._log.debug(
            f"Rendering template `{template_output_path}` with config `{template_config}`"
        )

        # if a template is not provided, find the template in the environment
        if not template:
            template = template_environment.get_template(template_name)

        if write_rendered_template:
            template.stream(**template_config).dump(str(template_output_path))
        else:
            template_stream = template.stream(**template_config)
            while next(template_stream, None):
                pass

    def _scaffold_directory(
        self,
        env: Environment,
        template_directory_prefix: str = "",
        overwrite_existing_files: bool = True,
    ):
        """
        Render all templates discovered under the root directory of the provided template environment.
        Rendered templates are output relative to their location in the template directory, prefixed
        with `<output directory>/template_directory_prefix`.
        Only files ending with `.j2` are rendered.
        """
        META_TEMPLATE_NAME = "__meta__.j2"
        # render the base templates
        templates = env.list_templates(extensions=["j2"])
        # if a module have a `__meta__.j2` file, then
        # that module is responsible for rendering its
        # own templates.
        # the template name will be without its parent directory
        # because the module renderer uses the module
        # template directory as the root directory.
        if META_TEMPLATE_NAME in templates:
            self._render_template(
                META_TEMPLATE_NAME,
                env,
                template_directory_prefix=template_directory_prefix,
                overwrite_existing_files=overwrite_existing_files,
                write_rendered_template=False,
            )
        else:
            for template_name in templates:
                self._render_template(
                    template_name,
                    env,
                    template_directory_prefix=template_directory_prefix,
                    overwrite_existing_files=overwrite_existing_files,
                )

    def _execute_module_hooks(self, module_template_directory: str):
        """
        Modules may have a file named `__hook__.py`. If it exists, it
        is executed as part of the scaffolding process.

        Hooks that can be configured in a module are:
            Called when an application is being created:
            `on_create(config: dict[str, Any], log: Logger) -> None`
        """
        module_hook_path = Path(module_template_directory, "__hook__.py")
        if self._provider.has_resource(str(module_hook_path)):
            # load the module from its path
            # and execute it
            spec = spec_from_file_location(
                "__hook__",
                self._provider.get_resource_filename(
                    self._manager, str(module_hook_path)
                ),
            )
            if spec is None or spec.loader is None:
                raise Exception(
                    f"Module cannot be created from path {module_hook_path}"
                )
            module = module_from_spec(spec)
            spec.loader.exec_module(module)

            for module_name, module_var in vars(module).items():
                if not module_name.startswith("on_"):
                    continue
                # Execute methods starting with `on_`.
                # Since `_execute_module_hooks` is currently
                # only called when an application is being created,
                # any such method is only called when an application
                # is being created. Although the only documented
                # method allowed is `on_create`, any such prefixed
                # method will be called.
                module_var(self._config_dict, self._log)

    def render_module_template(
        self,
        module: ScaffoldModule,
        module_template_directory: Path,
        overwrite_existing_files: bool = True,
    ):
        def _render_module_template(
            template_name: str,
            config: dict[Any, Any],
        ):
            meta_module_env = Environment(
                trim_blocks=True,
                lstrip_blocks=True,
                loader=PackageLoader(
                    "Ligare.web",
                    str(Path(module_template_directory, "templates")),
                ),
            )
            # set a test module-specific "meta" config for
            # use by whatever templates this loop will render
            self._config_dict["meta"] = config

            self._render_template(
                template_name,
                meta_module_env,
                template_directory_prefix=f"{self._config.application.module_name}/modules/{module.module_name}",
                overwrite_existing_files=overwrite_existing_files,
            )

        return _render_module_template

    def _scaffold_modules(self, overwrite_existing_files: bool = True):
        """
        Render any modules configured to render.
        """
        if self._config.modules is None:
            self._log.debug("No optional modules to scaffold.")
            return

        # each module's configuration currently only includes
        # the module's name. This name also matches the filesystem
        # directory for the module's templates.
        for module in self._config.modules:
            # module templates are stored under `optional/`
            # because we don't always want to render any given module.
            module_template_directory = Path(
                f"scaffolding/templates/optional/{{{{application.module_name}}}}/modules/{module.module_name}"
            )
            # executed module hooks before any rendering is done
            # so the hooks can modify the config or do other
            # work if it's needed.
            self._execute_module_hooks(str(module_template_directory))

            # all modules contain templates under templates/
            # but only modules with __meta__.j2 control how
            # their own templates are rendered.
            if self._provider.has_resource(
                str(Path(module_template_directory, "__meta__.j2"))
            ):
                module_env = Environment(
                    trim_blocks=True,
                    lstrip_blocks=True,
                    loader=PackageLoader(
                        "Ligare.web", str(Path(module_template_directory))
                    ),
                )
            else:
                module_env = Environment(
                    trim_blocks=True,
                    lstrip_blocks=True,
                    loader=PackageLoader(
                        "Ligare.web",
                        str(Path(module_template_directory, "templates")),
                    ),
                )

            # this is necessary to overwrite for every module
            # because the directory and module config are
            # unique for each one.
            module_env.globals["render_module_template"] = self.render_module_template(  # pyright: ignore[reportArgumentType]
                module=module,
                module_template_directory=module_template_directory,
                overwrite_existing_files=overwrite_existing_files,
            )

            self._scaffold_directory(
                module_env,
                # prefix the directory so that rendered templates
                # are output _not_ relative to the template environment
                # root, which would store the rendered templates in `./`.
                template_directory_prefix=f"{self._config.application.module_name}/modules/{module.module_name}",
                overwrite_existing_files=overwrite_existing_files,
            )

    def _scaffold_endpoints(
        self, env: Environment, overwrite_existing_files: bool = True
    ):
        """
        Render API endpoint templates.
        """
        # {{application.module_name}}/endpoints/{{operation.module_name}}.py.j2 is the file name - this is _not_ meant to be an interpolated string
        endpoint_template_filename = (
            "{{application.module_name}}/endpoints/{{operation.module_name}}.py.j2"
        )
        # technically, `self._config.endpoints` is always set in `__main__`
        # but we do this check to satisfy the type checker.
        if self._config.endpoints is None:
            self._log.debug("No endpoints to scaffold.")
            return

        # render optional API endpoint templates.
        # these templates are stored under `optional/` because:
        # 1. they are functionally optional. an application can be
        #    scaffolded without any endpoints, although `__main__` forces
        #    at least one endpoint (defaulting to the application name).
        # 2. more than one endpoint can be rendered, so a location
        #    for the templates that is not used to render an entire
        #    directory of templates (like the `base` templates) is needed.
        for endpoint in [asdict(_dataclass) for _dataclass in self._config.endpoints]:
            template = env.get_template(endpoint_template_filename)

            if template.name is None:
                self._log.error(
                    f"Could not find template `{endpoint_template_filename}`."
                )
                continue

            template_string_config = merge(self._config_dict.copy(), endpoint)

            # render the template output path, replacing the jinja2
            # directives with their associated values from `self._config_dict`.
            rendered_template_name = self._render_template_string(
                template.name, template_string_config
            )

            operation = endpoint["operation"]
            self._log.info(
                f'"{operation["url_path_name"]}" will be accessible at {endpoint["hostname"]}/{operation["url_path_name"]}'
            )

            # make the current endpoint configuration available
            # in a top-level `endpoint` variable
            template_config = self._config_dict.copy()
            template_config["endpoint"] = endpoint

            self._render_template(
                rendered_template_name,
                env,
                template_config=template_config,
                template=template,
                overwrite_existing_files=overwrite_existing_files,
            )

    # it is safe for create if:
    #      - the current directory is not a scaffolded application
    # it is safe for modify if:
    #      - the <application name> directory exists
    #      - the <application name> directory is a scaffolded application
    #      - the scaffolder is run from the scaffolded application's parent directory
    def _check_scaffolded_application_exists(self, test_path: Path):
        """
        Determine whether the application <config.application.module_name> is an importable
        Ligare.web application that has been previously scaffolded with
        this tool.

        :param test_path: The directory to test for the existence of a scaffolded application.
        """
        application_import_path = Path(".", test_path)
        self._log.debug(
            f"Adding `{application_import_path}` to process module import path."
        )
        # modifying the import path to include a relative directory
        # is less than ideal and potentionally dangerous. we should
        # consider forking here or doing something else to prevent
        # the global modification.
        sys.path.insert(0, str(application_import_path))

        try:
            # if any of these steps fail, then the `test_path`
            # does not already contain a scaffolded application
            loader = SourceFileLoader(
                self._config.application.module_name,
                str(
                    Path(
                        test_path,
                        self._config.application.module_name,
                        "__init__.py",
                    )
                ),
            )
            mod = types.ModuleType(loader.name)
            sys.modules[self._config.application.module_name] = mod
            module = loader.load_module()
            _ = module.__version__
            _ = module._version.__ligare_scaffold__
            # if no errors occur, then the `test_path`
            # contains a scaffolded application
            return True
        except Exception as e:
            self._log.debug(str(e), exc_info=True)
            return False
        finally:
            if self._config.application.module_name in sys.modules:
                del sys.modules[self._config.application.module_name]
            popped_import_path = sys.path.pop(0)
            self._log.debug(
                f"Popped `{popped_import_path}` from process module import path. Expected to pop `{application_import_path}`."
            )

    def scaffold(self):
        in_parent_directory = self._check_scaffolded_application_exists(
            Path(self._config.output_directory)
        )
        in_application_directory = self._check_scaffolded_application_exists(Path("."))

        # "create" can run from any directory that is not an existing
        # application's root directory.
        if self._config.mode == "create" and in_application_directory:
            self._log.critical(
                "Attmpted to scaffold a new application in the same directory as an existing application. This is not supported. Change your working directory to the application's parent directory, or run this from a directory that does not contain an existing application."
            )
            return

        # modify can only run from an existing application's
        # parent directory.
        if self._config.mode == "modify" and not in_parent_directory:
            self._log.critical(
                "Attempted to modify an existing application from a directory that is not the existing application's parent directory. This is not supported. Change your working directory to the application's parent directory."
            )
            return

        if (
            # Only in create mode do we make absolutely certain that
            # the user is intentionally using a directory that
            # already exists, if it indeed does.
            self._config.mode == "create"
            and Path(self._config.output_directory).exists()
        ):
            self._log.debug(
                f"The directory `{self._config.output_directory}` already exists. Prompting the user whether to continue."
            )
            response = input(
                f"\nThe directory `{self._config.output_directory}` already exists. It is likely that files will be overwritten.\nDo you want to continue? [y/N] "
            )
            if not response.upper() == "Y":
                self._log.debug(f"User's input is `{response}`. Not continuing.")
                return
            self._log.debug(f"User's input is `{response}`. Continuing.")

        # used for the primary set of templates that a
        # scaffolded application is made up of.
        base_env = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=PackageLoader("Ligare.web", "scaffolding/templates/base"),
        )
        # used for the selected template type templates
        # that can replace files from the base templates.
        template_type_env = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=PackageLoader(
                "Ligare.web", f"scaffolding/templates/{self._config.template_type}"
            ),
        )
        # used for templates under `optional/`
        optional_env = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=PackageLoader("Ligare.web", "scaffolding/templates/optional"),
        )

        if self._config.mode == "create":
            # scaffold modules first so they can alter the config if necessary.
            # a template environment is not passed in because `scaffold_modules`
            # creates a new environment for each module.
            self._scaffold_modules()

            self._scaffold_directory(base_env)
            # template type should go after base so it can override any base templates.
            self._scaffold_directory(template_type_env)

            self._scaffold_endpoints(optional_env)
        # other mode is "modify"
        else:
            # currently only endpoints can be modified through
            # this tool, and they can only be added.
            self._scaffold_endpoints(optional_env, overwrite_existing_files=False)



"""
This type stub file was generated by pyright.
"""

import starlette.requests
import starlette.responses
from json_logging.framework_base import AppRequestInstrumentationConfigurator, RequestAdapter, ResponseAdapter
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

class JSONLoggingASGIMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, exclude_url_patterns=...) -> None:
        ...
    
    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        ...
    


class FastAPIAppRequestInstrumentationConfigurator(AppRequestInstrumentationConfigurator):
    def config(self, app, exclude_url_patterns=...): # -> None:
        ...
    


class FastAPIRequestAdapter(RequestAdapter):
    @staticmethod
    def get_request_class_type(): # -> type[Request]:
        ...
    
    @staticmethod
    def support_global_request_object(): # -> Literal[False]:
        ...
    
    @staticmethod
    def get_current_request():
        ...
    
    def get_remote_user(self, request: starlette.requests.Request): # -> Any | Literal['-']:
        ...
    
    def get_http_header(self, request: starlette.requests.Request, header_name, default=...): # -> str | None:
        ...
    
    def set_correlation_id(self, request_, value): # -> None:
        ...
    
    def get_correlation_id_in_request_context(self, request: starlette.requests.Request): # -> Any | None:
        ...
    
    def get_protocol(self, request: starlette.requests.Request): # -> str:
        ...
    
    def get_path(self, request: starlette.requests.Request): # -> str:
        ...
    
    def get_content_length(self, request: starlette.requests.Request): # -> str:
        ...
    
    def get_method(self, request: starlette.requests.Request): # -> str:
        ...
    
    def get_remote_ip(self, request: starlette.requests.Request): # -> str:
        ...
    
    def get_remote_port(self, request: starlette.requests.Request): # -> int:
        ...
    


class FastAPIResponseAdapter(ResponseAdapter):
    def get_status_code(self, response: starlette.responses.Response): # -> int:
        ...
    
    def get_response_size(self, response: starlette.responses.Response): # -> str:
        ...
    
    def get_content_type(self, response: starlette.responses.Response): # -> str:
        ...
    





import uuid
from typing import Callable, Literal
from uuid import uuid4

import pytest
from flask import Flask, Response, abort
from Ligare.web.config import Config
from Ligare.web.middleware import bind_errorhandler
from Ligare.web.middleware.consts import CORRELATION_ID_HEADER
from Ligare.web.middleware.flask import (
    _get_correlation_id,  # pyright: ignore[reportPrivateUsage]
)
from Ligare.web.middleware.flask import bind_requesthandler
from Ligare.web.testing.create_app import (
    CreateFlaskApp,
    FlaskClientInjector,
    FlaskClientInjectorConfigurable,
    RequestConfigurable,
)
from mock import MagicMock
from pytest_mock import MockerFixture
from werkzeug.exceptions import BadRequest, HTTPException, Unauthorized


class TestFlaskMiddleware(CreateFlaskApp):
    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___register_api_response_handlers__sets_correlation_id_response_header_when_not_set_in_request_header(
        self,
        format: Literal["plaintext", "JSON"],
        flask_client_configurable: FlaskClientInjectorConfigurable,
        basic_config: Config,
    ):
        basic_config.logging.format = format
        flask_client = next(flask_client_configurable(basic_config))
        response = flask_client.client.get("/")

        assert response.headers[CORRELATION_ID_HEADER]
        _ = uuid.UUID(response.headers[CORRELATION_ID_HEADER])

    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___register_api_response_handlers__sets_correlation_id_response_header_when_set_in_request_header(
        self,
        format: Literal["plaintext", "JSON"],
        flask_client_configurable: FlaskClientInjectorConfigurable,
        basic_config: Config,
    ):
        basic_config.logging.format = format
        flask_client = next(flask_client_configurable(basic_config))
        correlation_id = str(uuid4())
        response = flask_client.client.get(
            "/", headers={CORRELATION_ID_HEADER: correlation_id}
        )

        assert response.headers[CORRELATION_ID_HEADER] == correlation_id

    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___get_correlation_id__validates_correlation_id_when_set_in_request_headers(
        self,
        format: Literal["plaintext", "JSON"],
        flask_request_configurable: RequestConfigurable,
        basic_config: Config,
    ):
        basic_config.logging.format = format
        correlation_id = "abc123"
        with flask_request_configurable(
            basic_config, {"headers": {CORRELATION_ID_HEADER: correlation_id}}
        ):
            with pytest.raises(
                ValueError, match="^badly formed hexadecimal UUID string$"
            ):
                _ = _get_correlation_id(MagicMock())

    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___get_correlation_id__uses_existing_correlation_id_when_set_in_request_headers(
        self,
        format: Literal["plaintext", "JSON"],
        flask_request_configurable: RequestConfigurable,
        basic_config: Config,
    ):
        basic_config.logging.format = format
        correlation_id = str(uuid4())
        with flask_request_configurable(
            basic_config, {"headers": {CORRELATION_ID_HEADER: correlation_id}}
        ):
            returned_correlation_id = _get_correlation_id(MagicMock())
            assert correlation_id == returned_correlation_id

    @pytest.mark.parametrize("format", ["plaintext", "JSON"])
    def test___get_correlation_id__sets_correlation_id(
        self,
        format: Literal["plaintext", "JSON"],
        flask_request_configurable: RequestConfigurable,
        basic_config: Config,
    ):
        basic_config.logging.format = format
        with flask_request_configurable(basic_config):
            correlation_id = _get_correlation_id(MagicMock())

            assert correlation_id
            _ = uuid.UUID(correlation_id)

    def test__bind_requesthandler__returns_decorated_flask_request_hook(
        self,
        flask_client: FlaskClientInjector,
    ):
        flask_request_hook_mock = MagicMock()

        wrapped_decorator = bind_requesthandler(
            flask_client.client.application, flask_request_hook_mock
        )
        _ = wrapped_decorator(lambda: None)

        assert flask_request_hook_mock.called

    def test__bind_requesthandler__calls_decorated_function_when_app_is_run(
        self,
        flask_client: FlaskClientInjector,
    ):
        wrapped_handler_decorator = bind_requesthandler(
            flask_client.client.application, Flask.before_request
        )
        request_handler_mock = MagicMock()
        _ = wrapped_handler_decorator(request_handler_mock)

        _ = flask_client.client.get("/")

        assert request_handler_mock.called

    @pytest.mark.parametrize("code_or_exception", [Exception, HTTPException, 401])
    def test__bind_errorhandler__binds_flask_errorhandler(
        self,
        code_or_exception: type[Exception] | int,
        flask_client: FlaskClientInjector,
        mocker: MockerFixture,
    ):
        flask_errorhandler_mock = mocker.patch("flask.Flask.errorhandler")

        _ = bind_errorhandler(flask_client.client.application, code_or_exception)

        flask_errorhandler_mock.assert_called_with(code_or_exception)

    @pytest.mark.parametrize(
        "code_or_exception_type,expected_exception_type,failure_lambda",
        [
            (
                Exception,
                ZeroDivisionError,
                lambda: 1 / 0,  # 1/0 to raise an exception (any exception)
            ),
            (HTTPException, BadRequest, lambda: abort(400)),
            (401, Unauthorized, lambda: abort(401)),
        ],
    )
    def test__bind_errorhandler__calls_decorated_function_with_correct_error_when_error_occurs_during_request(
        self,
        code_or_exception_type: type[Exception] | int,
        expected_exception_type: type[Exception],
        failure_lambda: Callable[[], Response],
    ):
        application_errorhandler_mock = MagicMock()
        with Flask("foo").test_client() as test_client:
            _ = bind_errorhandler(test_client.application, code_or_exception_type)(
                application_errorhandler_mock
            )

            _ = test_client.application.route("/")(failure_lambda)

            _ = test_client.get("/")

        assert application_errorhandler_mock.called
        assert isinstance(
            application_errorhandler_mock.call_args[0][0], expected_exception_type
        )



"""
This type stub file was generated by pyright.
"""

from starlette.types import Receive, Scope, Send

"""
This module defines a Validator interface with base functionality that can be subclassed
for custom validators provided to the RequestValidationMiddleware.
"""
class AbstractRequestBodyValidator:
    """
    Validator interface with base functionality that can be subclassed for custom validators.

    .. note: Validators load the whole body into memory, which can be a problem for large payloads.
    """
    MUTABLE_VALIDATION = ...
    MAX_MESSAGE_LENGTH = ...
    def __init__(self, *, schema: dict, required: bool = ..., nullable: bool = ..., encoding: str, strict_validation: bool, **kwargs) -> None:
        """
        :param schema: Schema of operation to validate
        :param required: Whether RequestBody is required
        :param nullable: Whether RequestBody is nullable
        :param encoding: Encoding of body (passed via Content-Type header)
        :param kwargs: Additional arguments for subclasses
        :param strict_validation: Whether to allow parameters not defined in the spec
        """
        ...
    
    async def wrap_receive(self, receive: Receive, *, scope: Scope) -> Receive:
        """
        Wrap the provided `receive` channel with request body validation.

        This method updates the provided `scope` in place with the right `Content-Length` header.
        """
        ...
    


class AbstractResponseBodyValidator:
    """
    Validator interface with base functionality that can be subclassed for custom validators.

    .. note: Validators load the whole body into memory, which can be a problem for large payloads.
    """
    def __init__(self, scope: Scope, *, schema: dict, nullable: bool = ..., encoding: str) -> None:
        ...
    
    def wrap_send(self, send: Send) -> Send:
        """Wrap the provided send channel with response body validation"""
        ...
    





"""
This type stub file was generated by pyright.
"""

import abc
import typing as t
from collections.abc import Mapping
from jsonschema import Draft4Validator
from .operations import AbstractOperation, OpenAPIOperation, Swagger2Operation

"""
This module defines Python interfaces for OpenAPI specifications.
"""
validate_properties = Draft4Validator.VALIDATORS["properties"]
def create_spec_validator(spec: dict) -> Draft4Validator:
    """Create a Validator to validate an OpenAPI spec against the OpenAPI schema.

    :param spec: specification to validate
    """
    ...

NO_SPEC_VERSION_ERR_MSG = ...
def canonical_base_path(base_path):
    """
    Make given "basePath" a canonical base URL which can be prepended to paths starting with "/".
    """
    ...

class Specification(Mapping):
    operation_cls: t.Type[AbstractOperation]
    def __init__(self, raw_spec, *, base_uri=...) -> None:
        ...
    
    def get_path_params(self, path): # -> list[Any]:
        ...
    
    def get_operation(self, path, method): # -> Mapping[Any, Any] | list[Any] | tuple[Any, ...]:
        ...
    
    @property
    def raw(self): # -> Any:
        ...
    
    @property
    def version(self): # -> tuple[int, ...]:
        ...
    
    @property
    def security(self): # -> None:
        ...
    
    @property
    @abc.abstractmethod
    def security_schemes(self):
        ...
    
    def __getitem__(self, k):
        ...
    
    def __iter__(self): # -> Iterator[Any]:
        ...
    
    def __len__(self): # -> int:
        ...
    
    @classmethod
    def from_file(cls, spec, *, arguments=..., base_uri=...): # -> Swagger2Specification | OpenAPISpecification:
        """
        Takes in a path to a YAML file, and returns a Specification
        """
        ...
    
    @classmethod
    def from_dict(cls, spec, *, base_uri=...): # -> Swagger2Specification | OpenAPISpecification:
        """
        Takes in a dictionary, and returns a Specification
        """
        ...
    
    def clone(self): # -> Self:
        ...
    
    @classmethod
    def load(cls, spec, *, arguments=...): # -> Swagger2Specification | OpenAPISpecification:
        ...
    
    def with_base_path(self, base_path): # -> Self:
        ...
    
    @property
    @abc.abstractmethod
    def base_path(self): # -> None:
        ...
    
    @base_path.setter
    @abc.abstractmethod
    def base_path(self, base_path): # -> None:
        ...
    


class Swagger2Specification(Specification):
    """Python interface for a Swagger 2 specification."""
    yaml_name = ...
    operation_cls = Swagger2Operation
    openapi_schema = ...
    @property
    def produces(self):
        ...
    
    @property
    def consumes(self):
        ...
    
    @property
    def definitions(self):
        ...
    
    @property
    def parameter_definitions(self):
        ...
    
    @property
    def response_definitions(self):
        ...
    
    @property
    def security_schemes(self): # -> dict[Any, Any]:
        ...
    
    @property
    def base_path(self): # -> str:
        ...
    
    @base_path.setter
    def base_path(self, base_path): # -> None:
        ...
    


class OpenAPISpecification(Specification):
    """Python interface for an OpenAPI 3 specification."""
    yaml_name = ...
    operation_cls = OpenAPIOperation
    openapi_schema = ...
    @property
    def security_schemes(self):
        ...
    
    @property
    def components(self):
        ...
    
    @property
    def base_path(self): # -> LiteralString:
        ...
    
    @base_path.setter
    def base_path(self, base_path): # -> None:
        ...
    





"""
This type stub file was generated by pyright.
"""

import abc
import typing as t
from starlette.types import ASGIApp, Receive, Scope, Send
from connexion.operations import AbstractOperation
from connexion.resolver import Resolver
from connexion.spec import Specification

logger = ...
ROUTING_CONTEXT = ...
class SpecMiddleware(abc.ABC):
    """Middlewares that need the specification(s) to be registered on them should inherit from this
    base class"""
    @abc.abstractmethod
    def add_api(self, specification: Specification, **kwargs) -> t.Any:
        """
        Register an API represented by a single OpenAPI specification on this middleware.
        Multiple APIs can be registered on a single middleware.
        """
        ...
    
    @abc.abstractmethod
    async def __call__(self, scope: Scope, receive: Receive, send: Send): # -> None:
        ...
    


class AbstractSpecAPI:
    """Base API class with only minimal behavior related to the specification."""
    def __init__(self, specification: Specification, base_path: t.Optional[str] = ..., resolver: t.Optional[Resolver] = ..., uri_parser_class=..., *args, **kwargs) -> None:
        ...
    


OP = t.TypeVar("OP")
class AbstractRoutingAPI(AbstractSpecAPI, t.Generic[OP]):
    """Base API class with shared functionality related to routing."""
    def __init__(self, *args, pythonic_params=..., resolver_error_handler: t.Optional[t.Callable] = ..., **kwargs) -> None:
        ...
    
    def add_paths(self, paths: t.Optional[dict] = ...) -> None:
        """
        Adds the paths defined in the specification as operations.
        """
        ...
    
    def add_operation(self, path: str, method: str) -> None:
        """
        Adds one operation to the api.

        This method uses the OperationID identify the module and function that will handle the operation

        From Swagger Specification:

        **OperationID**

        A friendly name for the operation. The id MUST be unique among all operations described in the API.
        Tools and libraries MAY use the operation id to uniquely identify an operation.
        """
        ...
    
    @abc.abstractmethod
    def make_operation(self, operation: AbstractOperation) -> OP:
        """Build an operation to register on the API."""
        ...
    


class RoutedAPI(AbstractSpecAPI, t.Generic[OP]):
    def __init__(self, specification: Specification, *args, next_app: ASGIApp, **kwargs) -> None:
        ...
    
    def add_paths(self) -> None:
        ...
    
    def add_operation(self, path: str, method: str) -> None:
        ...
    
    @abc.abstractmethod
    def make_operation(self, operation: AbstractOperation) -> OP:
        """Create an operation of the `operation_cls` type."""
        ...
    


API = t.TypeVar("API", bound="RoutedAPI")
class RoutedMiddleware(SpecMiddleware, t.Generic[API]):
    """Baseclass for middleware that wants to leverage the RoutingMiddleware to route requests to
    its operations.

    The RoutingMiddleware adds the operation_id to the ASGI scope. This middleware registers its
    operations by operation_id at startup. At request time, the operation is fetched by an
    operation_id lookup.
    """
    api_cls: t.Type[API]
    def __init__(self, app: ASGIApp) -> None:
        ...
    
    def add_api(self, specification: Specification, **kwargs) -> API:
        ...
    
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Fetches the operation related to the request and calls it."""
        ...
    


class MissingOperation(Exception):
    """Missing operation"""
    ...





"""
This type stub file was generated by pyright.
"""

from connexion.datastructures import MediaTypeDict

from .abstract import AbstractRequestBodyValidator, AbstractResponseBodyValidator
from .form_data import FormDataValidator, MultiPartFormDataValidator
from .json import (
    DefaultsJSONRequestBodyValidator,
    JSONRequestBodyValidator,
    JSONResponseBodyValidator,
    TextResponseBodyValidator,
)
from .parameter import ParameterValidator

VALIDATOR_MAP = ...

__all__ = (
    "MediaTypeDict",
    "AbstractRequestBodyValidator",
    "AbstractResponseBodyValidator",
    "FormDataValidator",
    "MultiPartFormDataValidator",
    "DefaultsJSONRequestBodyValidator",
    "JSONRequestBodyValidator",
    "JSONResponseBodyValidator",
    "TextResponseBodyValidator",
    "ParameterValidator",
    "VALIDATOR_MAP",
)



"""
This type stub file was generated by pyright.
"""

"""
This module defines error handlers, operations that produce proper response problems.
"""
logger = ...
RESOLVER_ERROR_ENDPOINT_RANDOM_DIGITS = ...
class ResolverErrorHandler:
    """
    Handler for responding to ResolverError.
    """
    def __init__(self, status_code, exception) -> None:
        ...
    
    @property
    def function(self): # -> Callable[..., NoReturn]:
        ...
    
    def handle(self, *args, **kwargs):
        ...
    
    @property
    def operation_id(self): # -> Literal['noop']:
        ...
    
    @property
    def randomize_endpoint(self): # -> Literal[6]:
        ...
    
    def get_path_parameter_types(self): # -> dict[Any, Any]:
        ...
    
    @property
    def uri_parser_class(self): # -> Literal['dummy']:
        ...
    
    @property
    def api(self): # -> Literal['dummy']:
        ...
    
    def get_mimetype(self): # -> Literal['dummy']:
        ...
    
    async def __call__(self, *args, **kwargs):
        ...
    





import logging
import sys
from argparse import ArgumentParser, Namespace
from functools import partial
from os import environ
from pathlib import Path
from typing import Callable, Literal, NamedTuple

from Ligare.programming.cli.argparse import DisallowDuplicateValues, disallow
from Ligare.web.scaffolding import (
    Operation,
    ScaffoldConfig,
    ScaffoldEndpoint,
    Scaffolder,
    ScaffoldModule,
)
from typing_extensions import final

APPLICATION_ENDPOINT_PATH_NAME = "application"


class ScaffoldInputArgs(Namespace):
    mode: Literal["create", "modify"]  # pyright: ignore[reportUninitializedInstanceVariable]
    mode_executor: "Callable[[ScaffoldParsedArgs], None]"  # pyright: ignore[reportUninitializedInstanceVariable]
    name: Operation  # pyright: ignore[reportUninitializedInstanceVariable]
    endpoints: list[Operation] | None = None
    template_type: Literal["basic", "openapi"] = "basic"
    modules: list[Literal["database", "test"]] | None = None
    output_directory: str | None = None


class ScaffoldParsedArgs(NamedTuple):
    mode: Literal["create", "modify"]
    mode_executor: "Callable[[ScaffoldParsedArgs], None]"
    name: Operation
    endpoints: list[Operation]
    template_type: Literal["basic", "openapi"]
    modules: list[Literal["database", "test"]] | None
    output_directory: str


@final
class ScaffolderCli:
    _log: logging.Logger

    def __init__(self, log_level: int | str | None = None) -> None:
        _log_level = log_level
        if log_level is not None:
            # support both integer and named log level values
            # from the envvar, e.g., `10` or `DEBUG`
            try:
                _log_level = int(log_level)
            except ValueError as e:
                if not "invalid literal for int" in str(e):
                    raise
                _log_level = getattr(logging, str(log_level).upper())

        logging.basicConfig(
            level=_log_level if _log_level is not None else logging.INFO,
            format="[%(levelname)-7s] %(message)s",
        )

        self._log = logging.getLogger()

    def _parse_args(self, args: list[str]) -> ScaffoldParsedArgs:
        parser = ArgumentParser(description="Ligare Web Project Scaffolding Tool")

        subparsers = parser.add_subparsers(help="Select mode", required=True)

        create_parser = subparsers.add_parser(
            "create", help="Create a new Ligare web application"
        )
        _ = create_parser.add_argument(
            "-n",
            metavar="name",
            dest="name",
            required=True,
            type=disallow([APPLICATION_ENDPOINT_PATH_NAME], "name", Operation),
            help="The name of the application.",
        )
        _ = create_parser.add_argument(
            "-e",
            metavar="endpoint",
            dest="endpoints",
            type=disallow([APPLICATION_ENDPOINT_PATH_NAME], "endpoint", Operation),
            action=DisallowDuplicateValues,
            help="The name of an endpoint to scaffold. Can be specified more than once. If not specified, an endpoint sharing the name of the application will be scaffolded.",
        )
        template_types = ["basic", "openapi"]
        _ = create_parser.add_argument(
            "-t",
            choices=template_types,
            dest="template_type",
            type=str.lower,
            default="basic",
            help="The type of template to scaffold.",
        )
        modules = ["database", "test"]
        _ = create_parser.add_argument(
            "-m",
            choices=modules,
            action="append",
            dest="modules",
            type=str,
            help="An optional module to include in the application. Can be specified more than once.",
        )
        _ = create_parser.add_argument(
            "-o",
            metavar="output directory",
            dest="output_directory",
            type=str,
            help="The output directory. The default is a new directory sharing the name of the application.",
        )
        create_parser.set_defaults(
            mode_executor=partial(self._run_create_mode), mode="create"
        )

        modify_parser = subparsers.add_parser(
            "modify", help="Modify an existing Ligare web application"
        )
        _ = modify_parser.add_argument(
            "-n",
            metavar="name",
            dest="name",
            required=True,
            type=disallow([APPLICATION_ENDPOINT_PATH_NAME], "name", Operation),
            help="The name of the application.",
        )
        _ = modify_parser.add_argument(
            "-e",
            metavar="endpoint",
            dest="endpoints",
            type=disallow([APPLICATION_ENDPOINT_PATH_NAME], "endpoint", Operation),
            action=DisallowDuplicateValues,
            help="The name of an endpoint to scaffold. Can be specified more than once. If not specified, an endpoint sharing the name of the application will be scaffolded.",
        )
        _ = modify_parser.add_argument(
            "-o",
            metavar="output directory",
            dest="output_directory",
            type=str,
            help="The output directory. The default is a directory sharing the name of the application.",
        )
        modify_parser.set_defaults(
            mode_executor=partial(self._run_modify_mode), mode="modify"
        )

        _args = parser.parse_args(args, namespace=ScaffoldInputArgs)

        if _args.output_directory is None:
            _args.output_directory = _args.name.module_name

        if _args.endpoints is None:
            _args.endpoints = [_args.name]
        else:
            endpoints = {
                endpoint.url_path_name: endpoint for endpoint in _args.endpoints
            }
            if APPLICATION_ENDPOINT_PATH_NAME in endpoints:
                self._log.warning(
                    f'The endpoint name "{APPLICATION_ENDPOINT_PATH_NAME}" is reserved and will not be scaffolded.'
                )
                _args.endpoints.remove(endpoints[APPLICATION_ENDPOINT_PATH_NAME])

        return ScaffoldParsedArgs(
            _args.mode,
            _args.mode_executor,
            _args.name,
            _args.endpoints,
            _args.template_type,
            _args.modules,
            _args.output_directory,
        )

    def _run_create_mode(self, args: ScaffoldParsedArgs):
        self._log.info("Running create mode.")

        self._log.info(
            f'Creating new application "{args.name}" from {args.template_type} template ...'
        )

        # TODO consider pulling licenses from GitHub
        # https://docs.github.com/en/rest/licenses/licenses?apiVersion=2022-11-28#get-all-commonly-used-licenses
        # https://docs.github.com/en/rest/licenses/licenses?apiVersion=2022-11-28#get-a-license
        config = ScaffoldConfig(
            mode=args.mode,
            template_type=args.template_type,
            output_directory=args.output_directory,
            application=args.name,
            modules=(
                []
                if args.modules is None
                else [ScaffoldModule(module_name=module) for module in args.modules]
            ),
            endpoints=[
                ScaffoldEndpoint(operation=endpoint) for endpoint in args.endpoints
            ],
        )

        scaffolder = Scaffolder(config, self._log)
        scaffolder.scaffold()

    def _run_modify_mode(self, args: ScaffoldParsedArgs):
        self._log.info("Running modify mode.")

        config = ScaffoldConfig(
            mode=args.mode,
            output_directory=args.output_directory,
            application=args.name,
            endpoints=[
                ScaffoldEndpoint(operation=endpoint) for endpoint in args.endpoints
            ],
        )

        scaffolder = Scaffolder(config, self._log)
        scaffolder.scaffold()

    def run(self, argv: list[str]):
        args = self._parse_args(argv)

        print(
            f'Scaffolding application named "{args.name}" under directory `{Path(args.output_directory).absolute()}`.'
        )

        args.mode_executor(args)


def scaffold(argv: list[str] | None = None, log_level: int | str | None = None):
    if not log_level:
        log_level = environ.get("LOG_LEVEL")

    if not argv:
        argv = sys.argv[1:]

    cli = ScaffolderCli(log_level)
    cli.run(argv)
    print("Done.")



"""
This type stub file was generated by pyright.
"""

import typing as t

from starlette.types import ASGIApp, Receive, Scope, Send

Lifespan = t.Callable[[t.Any], t.AsyncContextManager[t.Any]]

class LifespanMiddleware:
    """
    Middleware that adds support for Starlette lifespan handlers
    (https://www.starlette.io/lifespan/).
    """

    def __init__(
        self, next_app: ASGIApp, *, lifespan: t.Optional[Lifespan]
    ) -> None: ...
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None: ...

__all__ = ("t", "ASGIApp", "Receive", "Scope", "Send", "Lifespan", "LifespanMiddleware")



"""
This type stub file was generated by pyright.
"""

import abc

"""
This module defines URIParsers which parse query and path parameters according to OpenAPI
serialization rules.
"""
logger = ...
QUERY_STRING_DELIMITERS = ...
class AbstractURIParser(metaclass=abc.ABCMeta):
    parsable_parameters = ...
    def __init__(self, param_defns, body_defn) -> None:
        """
        a URI parser is initialized with parameter definitions.
        When called with a request object, it handles array types in the URI
        both in the path and query according to the spec.
        Some examples include:
        - https://mysite.fake/in/path/1,2,3/            # path parameters
        - https://mysite.fake/?in_query=a,b,c           # simple query params
        - https://mysite.fake/?in_query=a|b|c           # various separators
        - https://mysite.fake/?in_query=a&in_query=b,c  # complex query params
        """
        ...
    
    @property
    @abc.abstractmethod
    def param_defns(self): # -> None:
        """
        returns the parameter definitions by name
        """
        ...
    
    @property
    @abc.abstractmethod
    def param_schemas(self): # -> None:
        """
        returns the parameter schemas by name
        """
        ...
    
    def __repr__(self): # -> str:
        """
        :rtype: str
        """
        ...
    
    @abc.abstractmethod
    def resolve_form(self, form_data): # -> None:
        """Resolve cases where form parameters are provided multiple times."""
        ...
    
    @abc.abstractmethod
    def resolve_query(self, query_data): # -> None:
        """Resolve cases where query parameters are provided multiple times."""
        ...
    
    @abc.abstractmethod
    def resolve_path(self, path): # -> None:
        """Resolve cases where path parameters include lists"""
        ...
    
    def resolve_params(self, params, _in): # -> dict[Any, Any]:
        """
        takes a dict of parameters, and resolves the values into
        the correct array type handling duplicate values, and splitting
        based on the collectionFormat defined in the spec.
        """
        ...
    


class OpenAPIURIParser(AbstractURIParser):
    style_defaults = ...
    @property
    def param_defns(self): # -> dict[Any, Any]:
        ...
    
    @property
    def form_defns(self): # -> dict[Any, Any]:
        ...
    
    @property
    def param_schemas(self): # -> dict[Any, Any]:
        ...
    
    def resolve_form(self, form_data):
        ...
    
    def resolve_query(self, query_data): # -> dict[Any, Any]:
        ...
    
    def resolve_path(self, path_data): # -> dict[Any, Any]:
        ...
    


class Swagger2URIParser(AbstractURIParser):
    """
    Adheres to the Swagger2 spec,
    Assumes that the last defined query parameter should be used.
    """
    parsable_parameters = ...
    @property
    def param_defns(self): # -> dict[Any, Any]:
        ...
    
    @property
    def param_schemas(self): # -> dict[Any, Any]:
        ...
    
    def resolve_form(self, form_data): # -> dict[Any, Any]:
        ...
    
    def resolve_query(self, query_data): # -> dict[Any, Any]:
        ...
    
    def resolve_path(self, path_data): # -> dict[Any, Any]:
        ...
    


class FirstValueURIParser(Swagger2URIParser):
    """
    Adheres to the Swagger2 spec
    Assumes that the first defined query parameter should be used
    """
    ...


class AlwaysMultiURIParser(Swagger2URIParser):
    """
    Does not adhere to the Swagger2 spec, but is backwards compatible with
    connexion behavior in version 1.4.2
    """
    ...





"""
This type stub file was generated by pyright.
"""

class RequestAdapter:
    """
        Helper class help to extract logging-relevant information from HTTP request object
    """
    def __new__(cls, *arg, **kwargs): # -> Self@RequestAdapter:
        ...
    
    @staticmethod
    def support_global_request_object():
        """
            whether current framework supports global request object like Flask
        """
        ...
    
    @staticmethod
    def get_current_request():
        """
            get current request, should only implement for framework that support global request object
        """
        ...
    
    @staticmethod
    def get_request_class_type():
        """
            class type of request object, only need to specify in case the framework dont support global request object
        """
        ...
    
    def get_http_header(self, request, header_name, default=...):
        """
        get HTTP header value given it value name

        :param request: request object
        :param header_name: name of header
        :param default: default value if header value is not present
        :return:
        """
        ...
    
    def get_remote_user(self, request):
        """

        :param request: request object
        """
        ...
    
    def set_correlation_id(self, request, value):
        """
        We can safely assume that request is valid request object.\n
        Set correlation to request context. e.g Flask.g in Flask
        Made sure that we can access it later from get_correlation_id_in_request_context given the same request.

        :param value: correlation id string
        :param request: request object
        """
        ...
    
    def get_correlation_id_in_request_context(self, request):
        """
        We can safely assume that request is valid request object.

        :param request: request object
        """
        ...
    
    def get_protocol(self, request):
        """
        We can safely assume that request is valid request object.\n
        Gets the request protocol (e.g. HTTP/1.1).

        :return: The request protocol or '-' if it cannot be determined
        """
        ...
    
    def get_path(self, request):
        """
        We can safely assume that request is valid request object.\n
        Gets the request path.

       :return: the request path (e.g. /index.html)
        """
        ...
    
    def get_content_length(self, request):
        """
        We can safely assume that request is valid request object.\n
        The content length of the request.

        :return: the content length of the request or '-' if it cannot be determined
        """
        ...
    
    def get_method(self, request):
        """
        We can safely assume that request is valid request object.\n
        Gets the request method (e.g. GET, POST, etc.).

        :return: The request method or '-' if it cannot be determined
        """
        ...
    
    def get_remote_ip(self, request):
        """
        We can safely assume that request is valid request object.\n
        Gets the remote ip of the request initiator.

        :return: An ip address or '-' if it cannot be determined
        """
        ...
    
    def get_remote_port(self, request):
        """
        We can safely assume that request is valid request object.\n
        Gets the remote port of the request initiator.

        :return: A port or '-' if it cannot be determined
        """
        ...
    


class ResponseAdapter:
    """
        Helper class help to extract logging-relevant information from HTTP response object
    """
    def __new__(cls, *arg, **kwargs): # -> Self@ResponseAdapter:
        ...
    
    def get_status_code(self, response):
        """
        get response's integer status code

        :param response: response object
        """
        ...
    
    def get_response_size(self, response):
        """
        get response's size in bytes

        :param response: response object
        """
        ...
    
    def get_content_type(self, response):
        """
        get response's MIME/media type

        :param response: response object
        """
        ...
    


class FrameworkConfigurator:
    """
       Class to perform logging configuration for given framework as needed, like disable built in request logging and other utils logging
    """
    def __new__(cls, *args, **kw): # -> Self@FrameworkConfigurator:
        ...
    
    def config(self):
        """
            app logging configuration logic
        """
        ...
    


class AppRequestInstrumentationConfigurator:
    """
        Class to perform request instrumentation logging configuration. Should at least contains:
        1- register before-request hook and create a RequestInfo object, store it to request context
        2- register after-request hook and update response to stored RequestInfo object
        3 - re-configure framework loggers.
        NOTE: logger that is used to emit request instrumentation logs will need to assign to **self.request_logger**
    """
    def __new__(cls, *args, **kw): # -> Self@AppRequestInstrumentationConfigurator:
        ...
    
    def config(self, app, exclude_url_patterns=...):
        """
        configuration logic

        :param app:
        """
        ...
    
    def get_request_logger(self):
        """
        get the current logger that is used to logger the request instrumentation information
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

import typing as t
from contextvars import ContextVar
from starlette.types import ASGIApp, Receive, Scope, Send
from connexion.middleware import SpecMiddleware
from connexion.middleware.abstract import AbstractSpecAPI
from connexion.options import SwaggerUIOptions
from connexion.spec import Specification

logger = ...
_original_scope: ContextVar[Scope] = ...
class SwaggerUIAPI(AbstractSpecAPI):
    def __init__(self, *args, default: ASGIApp, swagger_ui_options: t.Optional[SwaggerUIOptions] = ..., **kwargs) -> None:
        ...
    
    @staticmethod
    def normalize_string(string): # -> str:
        ...
    
    def add_openapi_json(self): # -> None:
        """
        Adds openapi json to {base_path}/openapi.json
             (or {base_path}/swagger.json for swagger2)
        """
        ...
    
    def add_openapi_yaml(self): # -> None:
        """
        Adds openapi json to {base_path}/openapi.json
             (or {base_path}/swagger.json for swagger2)
        """
        ...
    
    def add_swagger_ui(self): # -> None:
        """
        Adds swagger ui to {base_path}/ui/
        """
        ...
    


class SwaggerUIMiddleware(SpecMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        """Middleware that hosts a swagger UI.

        :param app: app to wrap in middleware.
        """
        ...
    
    def add_api(self, specification: Specification, base_path: t.Optional[str] = ..., arguments: t.Optional[dict] = ..., **kwargs) -> None:
        """Add an API to the router based on a OpenAPI spec.

        :param specification: OpenAPI spec.
        :param base_path: Base path where to add this API.
        :param arguments: Jinja arguments to replace in the spec.
        """
        ...
    
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        ...
    
    async def default_fn(self, _scope: Scope, receive: Receive, send: Send) -> None:
        """
        Callback to call next app as default when no matching route is found.

        Unfortunately we cannot just pass the next app as default, since the router manipulates
        the scope when descending into mounts, losing information about the base path. Therefore,
        we use the original scope instead.

        This is caused by https://github.com/encode/starlette/issues/1336.
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

import argparse
import typing as t
from connexion.apps import AbstractApp

"""
This module defines a command-line interface (CLI) that runs an OpenAPI specification to be a
starting point for developing your API with Connexion.
"""
logger = ...
FLASK_APP = ...
ASYNC_APP = ...
AVAILABLE_APPS = ...
def run(app: AbstractApp, args: argparse.Namespace): # -> None:
    ...

parser = ...
subparsers = ...
run_parser = ...
def create_app(args: t.Optional[argparse.Namespace] = ...) -> AbstractApp:
    """Runs a server compliant with a OpenAPI/Swagger Specification file."""
    ...

def main(argv: t.Optional[t.List[str]] = ...) -> None:
    ...




"""
This type stub file was generated by pyright.
"""

import typing as t

"""
This module contains resolvers, functions that resolves the user defined view functions
from the operations defined in the OpenAPI spec.
"""
logger = ...
class Resolution:
    def __init__(self, function, operation_id) -> None:
        """
        Represents the result of operation resolution

        :param function: The endpoint function
        :type function: types.FunctionType
        """
        ...
    


class Resolver:
    def __init__(self, function_resolver: t.Callable = ...) -> None:
        """
        Standard resolver

        :param function_resolver: Function that resolves functions using an operationId
        """
        ...
    
    def resolve(self, operation): # -> Resolution:
        """
        Default operation resolver

        :type operation: connexion.operations.AbstractOperation
        """
        ...
    
    def resolve_operation_id(self, operation): # -> str:
        """
        Default operationId resolver

        :type operation: connexion.operations.AbstractOperation
        """
        ...
    
    def resolve_function_from_operation_id(self, operation_id):
        """
        Invokes the function_resolver

        :type operation_id: str
        """
        ...
    


class RelativeResolver(Resolver):
    """
    Resolves endpoint functions relative to a given root path or module.
    """
    def __init__(self, root_path, function_resolver=...) -> None:
        """
        :param root_path: The root path relative to which an operationId is resolved.
            Can also be a module. Has the same effect as setting
            `x-swagger-router-controller` or `x-openapi-router-controller` equal to
            `root_path` for every operation individually.
        :type root_path: typing.Union[str, types.ModuleType]
        :param function_resolver: Function that resolves functions using an operationId
        :type function_resolver: types.FunctionType
        """
        ...
    
    def resolve_operation_id(self, operation): # -> str:
        """Resolves the operationId relative to the root path, unless
        x-swagger-router-controller or x-openapi-router-controller is specified.

        :param operation: The operation to resolve
        :type operation: connexion.operations.AbstractOperation
        """
        ...
    


class RestyResolver(Resolver):
    """
    Resolves endpoint functions using REST semantics (unless overridden by specifying operationId)
    """
    def __init__(self, default_module_name: str, *, collection_endpoint_name: str = ...) -> None:
        """
        :param default_module_name: Default module name for operations
        :param collection_endpoint_name: Name of function to resolve collection endpoints to
        """
        ...
    
    def resolve_operation_id(self, operation): # -> str:
        """
        Resolves the operationId using REST semantics unless explicitly configured in the spec

        :type operation: connexion.operations.AbstractOperation
        """
        ...
    
    def resolve_operation_id_using_rest_semantics(self, operation): # -> str:
        """
        Resolves the operationId using REST semantics

        :type operation: connexion.operations.AbstractOperation
        """
        ...
    


class MethodResolverBase(RestyResolver):
    """
    Resolves endpoint functions based on Flask's MethodView semantics, e.g.

    .. code-block:: yaml

        paths:
            /foo_bar:
                get:
                    # Implied function call: api.FooBarView().get

    .. code-block:: python

        class FooBarView(MethodView):
            def get(self):
                return ...
            def post(self):
                return ...

    """
    _class_arguments_type = t.Dict[str, t.Dict[str, t.Union[t.Iterable, t.Dict[str, t.Any]]]]
    def __init__(self, *args, class_arguments: _class_arguments_type = ..., **kwargs) -> None:
        """
        :param args: Arguments passed to :class:`~RestyResolver`
        :param class_arguments: Arguments to instantiate the View Class in the format below
        :param kwargs: Keywords arguments passed to :class:`~RestyResolver`

        .. code-block:: python

            {
              "ViewName": {
                "args": (positional arguments,)
                "kwargs": {
                  "keyword": "argument"
                }
              }
            }
        """
        ...
    
    def resolve_operation_id(self, operation): # -> str:
        """
        Resolves the operationId using REST semantics unless explicitly configured in the spec
        Once resolved with REST semantics the view_name is capitalised and has 'View' added
        to it so it now matches the Class names of the MethodView

        :type operation: connexion.operations.AbstractOperation
        """
        ...
    
    def resolve_function_from_operation_id(self, operation_id):
        """
        Invokes the function_resolver

        :type operation_id: str
        """
        ...
    
    def resolve_method_from_class(self, view_name, meth_name, view_cls):
        """
        Returns the view function for the given view class.
        """
        ...
    


class MethodResolver(MethodResolverBase):
    """
    A generic method resolver that instantiates a class and extracts the method
    from it, based on the operation id.
    """
    def resolve_method_from_class(self, view_name, meth_name, view_cls): # -> Any:
        ...
    


class MethodViewResolver(MethodResolverBase):
    """
    A specialized method resolver that works with flask's method views.
    It resolves the method by calling as_view on the class.
    """
    def __init__(self, *args, **kwargs) -> None:
        ...
    
    def resolve_method_from_class(self, view_name, meth_name, view_cls):
        ...
    





"""
This type stub file was generated by pyright.
"""

import typing as t

"""
This module defines a SecurityHandlerFactory which supports the creation of
SecurityHandler instances for different security schemes.

It also exposes a `SECURITY_HANDLERS` dictionary which maps security scheme
types to SecurityHandler classes. This dictionary can be used to register
custom SecurityHandler classes for custom security schemes, or to overwrite
existing SecurityHandler classes.
This can be done by supplying a value for `security_map` argument of the
SecurityHandlerFactory.

Swagger 2.0 lets you define the following authentication types for an API:

- Basic authentication
- API key (as a header or a query string parameter)
- OAuth 2 common flows (authorization code, implicit, resource owner password credentials, client credentials)


Changes from OpenAPI 2.0 to OpenAPI 3.0
If you used OpenAPI 2.0 before, here is a summary of changes to help you get started with OpenAPI 3.0:
- securityDefinitions were renamed to securitySchemes and moved inside components.
- type: basic was replaced with type: http and scheme: basic.
- The new type: http is an umbrella type for all HTTP security schemes, including Basic, Bearer and other,
and the scheme keyword indicates the scheme type.
- API keys can now be sent in: cookie.
- Added support for OpenID Connect Discovery (type: openIdConnect).
- OAuth 2 security schemes can now define multiple flows.
- OAuth 2 flows were renamed to match the OAuth 2 Specification: accessCode is now authorizationCode,
and application is now clientCredentials.


OpenAPI uses the term security scheme for authentication and authorization schemes.
OpenAPI 3.0 lets you describe APIs protected using the following security schemes:

- HTTP authentication schemes (they use the Authorization header):
    - Basic
    - Bearer
    - other HTTP schemes as defined by RFC 7235 and HTTP Authentication Scheme Registry
- API keys in headers, query string or cookies
    - Cookie authentication
- OAuth 2
- OpenID Connect Discovery

"""
logger = ...
NO_VALUE = ...
class AbstractSecurityHandler:
    required_scopes_kw = ...
    client = ...
    security_definition_key: str
    environ_key: str
    def get_fn(self, security_scheme, required_scopes): # -> Callable[..., Coroutine[Any, Any, object | Any]] | None:
        """Returns the handler function"""
        ...
    
    @staticmethod
    def get_auth_header_value(request): # -> tuple[None, None] | tuple[Any, Any]:
        """
        Return Authorization type and value if any.
        If not Authorization, return (None, None)
        Raise OAuthProblem for invalid Authorization header
        """
        ...
    


class BasicSecurityHandler(AbstractSecurityHandler):
    """
    Security Handler for
    - `type: basic` (Swagger 2), and
    - `type: http` and `scheme: basic` (OpenAPI 3)
    """
    security_definition_key = ...
    environ_key = ...
    def check_basic_auth(self, basic_info_func): # -> Callable[..., Coroutine[Any, Any, object | Any]]:
        ...
    


class BearerSecurityHandler(AbstractSecurityHandler):
    """
    Security Handler for HTTP Bearer authentication.
    """
    security_definition_key = ...
    environ_key = ...
    def check_bearer_token(self, token_info_func): # -> Callable[..., Coroutine[Any, Any, object | Any]]:
        ...
    


class ApiKeySecurityHandler(AbstractSecurityHandler):
    """
    Security Handler for API Keys.
    """
    security_definition_key = ...
    environ_key = ...
    def get_fn(self, security_scheme, required_scopes): # -> Callable[..., object | Coroutine[Any, Any, object | Any]] | None:
        ...
    
    def check_api_key(self, api_key_info_func): # -> Callable[..., Coroutine[Any, Any, object | Any]]:
        ...
    
    @staticmethod
    def get_cookie_value(cookies, name): # -> str | None:
        """
        Returns cookie value by its name. `None` if no such value.

        :param cookies: str: cookies raw data
        :param name: str: cookies key
        """
        ...
    


class OAuthSecurityHandler(AbstractSecurityHandler):
    """
    Security Handler for the OAuth security scheme.
    """
    def get_fn(self, security_scheme, required_scopes): # -> Callable[..., object | Coroutine[Any, Any, object | Any]] | None:
        ...
    
    def get_tokeninfo_func(self, security_definition: dict) -> t.Optional[t.Callable]:
        """
        Gets the function for retrieving the token info.
        It is possible to specify a function or a URL. The function variant is
        preferred. If it is not found, the URL variant is used with the
        `get_token_info_remote` function.

        >>> get_tokeninfo_func({'x-tokenInfoFunc': 'foo.bar'})
        '<function foo.bar>'
        """
        ...
    
    @classmethod
    def get_scope_validate_func(cls, security_definition): # -> Any | Callable[..., Any] | None:
        """
        Gets the function for validating the token scopes.
        If it is not found, the default `validate_scope` function is used.

        >>> get_scope_validate_func({'x-scopeValidateFunc': 'foo.bar'})
        '<function foo.bar>'
        """
        ...
    
    @staticmethod
    def validate_scope(required_scopes, token_scopes): # -> bool:
        """
        :param required_scopes: Scopes required to access operation
        :param token_scopes: Scopes granted by authorization server
        :rtype: bool
        """
        ...
    
    def get_token_info_remote(self, token_info_url: str) -> t.Callable:
        """
        Return a function which will call `token_info_url` to retrieve token info.

        Returned function must accept oauth token in parameter.
        It must return a token_info dict in case of success, None otherwise.

        :param token_info_url: URL to get information about the token
        """
        ...
    
    def check_oauth_func(self, token_info_func, scope_validate_func): # -> Callable[..., Coroutine[Any, Any, object | Any]]:
        ...
    


SECURITY_HANDLERS = ...
class SecurityHandlerFactory:
    """
    A factory class for parsing security schemes and returning the appropriate
    security handler.

    By default, it will use the built-in security handlers specified in the
    SECURITY_HANDLERS dict, but you can also pass in your own security handlers
    to override the built-in ones.
    """
    def __init__(self, security_handlers: t.Optional[dict] = ...) -> None:
        ...
    
    def parse_security_scheme(self, security_scheme: dict, required_scopes: t.List[str]) -> t.Optional[t.Callable]:
        """Parses the security scheme and returns the function for verifying it.

        :param security_scheme: The security scheme from the spec.
        :param required_scopes: List of scopes for this security scheme.
        """
        ...
    
    @staticmethod
    async def security_passthrough(request):
        """Used when no security is required for the operation.

        Equivalent OpenAPI snippet:

        .. code-block:: yaml

            /helloworld
              get:
                security: []   # No security
                ...
        """
        ...
    
    @staticmethod
    def verify_none(request): # -> dict[Any, Any]:
        """Used for optional security.

        Equivalent OpenAPI snippet:

        .. code-block:: yaml

            security:
              - {}  # <--
              - myapikey: []
        """
        ...
    
    def verify_multiple_schemes(self, schemes): # -> Callable[..., Coroutine[Any, Any, object | dict[Any, Any]]]:
        """
        Verifies multiple authentication schemes in AND fashion.
        If any scheme fails, the entire authentication fails.

        :param schemes: mapping scheme_name to auth function
        :type schemes: dict
        :rtype: types.FunctionType
        """
        ...
    
    @classmethod
    def verify_security(cls, auth_funcs): # -> Callable[..., Coroutine[Any, Any, None]]:
        ...
    





import hashlib
from typing import TYPE_CHECKING, Any, Callable, Dict

from flask.json.tag import TaggedJSONSerializer
from flask.sessions import SessionMixin
from itsdangerous import URLSafeTimedSerializer

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer  # pragma: no cover

_EMPTY_SECRET_KEY_MESSAGE = (
    "It is insecure to use an empty `secret_key`. Refusing to continue."
)


def _get_serializer(
    secret_key: str,
    salt: str = "cookie-session",
    key_derivation: str = "hmac",
    digest_method: Callable[
        ["ReadableBuffer"], "hashlib._Hash"  # pyright: ignore[reportPrivateUsage]
    ] = hashlib.sha1,
):
    serializer = TaggedJSONSerializer()
    return URLSafeTimedSerializer(
        secret_key,
        salt=salt,
        serializer=serializer,
        signer_kwargs={
            "key_derivation": key_derivation,
            "digest_method": digest_method,
        },
    )


def decrypt_flask_cookie(secret_key: str, cookie_str: str) -> Dict[str, Any]:
    """
    Decrypt a Flask session cookie.
    https://gist.github.com/babldev/502364a3f7c9bafaa6db
    """
    if not secret_key:
        raise Exception(_EMPTY_SECRET_KEY_MESSAGE)

    serializer = _get_serializer(secret_key)
    data: Any = serializer.loads(cookie_str)

    if type(data) is not dict:
        raise AssertionError(
            f"Deserialized session data is not a dictionary. It is a `{type(data)}`."
        )

    return data  # pyright: ignore[reportUnknownVariableType]


def encrypt_flask_cookie(secret_key: str, data: Dict[str, Any] | SessionMixin) -> str:
    """
    Encrypt a dictionary into a Flask session cookie.
    """
    if not secret_key:
        raise Exception(_EMPTY_SECRET_KEY_MESSAGE)

    serializer = _get_serializer(secret_key)
    cookie_str = serializer.dumps(data)

    if type(cookie_str) is not str:
        raise AssertionError(
            f"Serialized cookie data is not a str. It is a `{type(cookie_str)}`."
        )

    return cookie_str



"""
This type stub file was generated by pyright.
"""

import logging
import json_logging
import json_logging.framework
import flask as flask
from json_logging.framework_base import AppRequestInstrumentationConfigurator, RequestAdapter, ResponseAdapter
from json_logging.util import is_not_match_any_pattern
from flask import request as request_obj

def is_flask_present(): # -> bool:
    ...

if is_flask_present():
    _current_request = ...
    _flask = flask
class FlaskAppRequestInstrumentationConfigurator(AppRequestInstrumentationConfigurator):
    def config(self, app, exclude_url_patterns=...): # -> None:
        ...
    


class FlaskRequestAdapter(RequestAdapter):
    @staticmethod
    def get_request_class_type():
        ...
    
    @staticmethod
    def support_global_request_object(): # -> Literal[True]:
        ...
    
    @staticmethod
    def get_current_request(): # -> Request:
        ...
    
    def get_remote_user(self, request): # -> Literal['-']:
        ...
    
    def get_http_header(self, request, header_name, default=...): # -> None:
        ...
    
    def set_correlation_id(self, request_, value): # -> None:
        ...
    
    def get_correlation_id_in_request_context(self, request): # -> Any | None:
        ...
    
    def get_protocol(self, request):
        ...
    
    def get_path(self, request):
        ...
    
    def get_content_length(self, request):
        ...
    
    def get_method(self, request):
        ...
    
    def get_remote_ip(self, request):
        ...
    
    def get_remote_port(self, request):
        ...
    


class FlaskResponseAdapter(ResponseAdapter):
    def get_status_code(self, response):
        ...
    
    def get_response_size(self, response):
        ...
    
    def get_content_type(self, response):
        ...
    





import logging
import logging as log
from abc import ABC
from dataclasses import dataclass
from functools import wraps
from logging import Logger
from typing import (
    Any,
    Callable,
    Dict,
    ParamSpec,
    Protocol,
    Sequence,
    TypedDict,
    TypeVar,
    cast,
    overload,
)
from urllib.parse import urlparse

import flask_login
from connexion import FlaskApp
from flask import (
    Blueprint,
    Flask,
    Response,
    current_app,
    redirect,
    request,
    session,
    url_for,
)
from flask.helpers import make_response
from flask.wrappers import Response
from flask_login import LoginManager as FlaskLoginManager
from flask_login import UserMixin as FlaskLoginUserMixin
from flask_login import login_user  # pyright: ignore[reportUnknownVariableType]
from flask_login import logout_user  # pyright: ignore[reportUnknownVariableType]
from flask_login import current_user
from flask_login import login_required as flask_login_required
from injector import Binder, Injector, inject
from Ligare.identity.config import Config, SAML2Config, SSOConfig
from Ligare.identity.dependency_injection import SAML2Module, SSOModule
from Ligare.identity.SAML2 import SAML2Client
from Ligare.platform.identity.user_loader import Role, UserId, UserLoader, UserMixin
from Ligare.programming.config import AbstractConfig
from Ligare.programming.patterns.dependency_injection import ConfigurableModule
from Ligare.web.config import Config
from Ligare.web.encryption import decrypt_flask_cookie
from saml2.validate import (
    MustValueError,
    NotValid,
    OutsideCardinality,
    ResponseLifetimeExceed,
    ShouldValueError,
    ToEarly,
)
from starlette.types import ASGIApp, Receive, Scope, Send
from typing_extensions import NotRequired, override
from werkzeug.exceptions import BadRequest, Forbidden, Unauthorized


class _LoginUserMixin(FlaskLoginUserMixin, ABC):
    """Used strictly for typecasting. This matches the Protocol UserMixin in user_loader."""

    id: UserId
    roles: Sequence[Role]


class AuthCheckUser(Protocol):
    id: UserId
    roles: Sequence[Role]


class AuthCheckOverrideCallable(Protocol):
    def __call__(self, user: AuthCheckUser, *args: Any, **kwargs: Any) -> bool: ...


P = ParamSpec("P")
R = TypeVar("R")


@overload
def login_required() -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Require a user session for the decorated API. No further requirements are applied.
    In effect, this uses `flask_login.login_required` and is used in its usual way.

    This is meant to be used as a decorator with `@login_required`. It is the the equivalent
    of using the decorator `@flask_login.login_required`.

    :return Callable[[Callable[P, R]], Callable[P, R]]: Returns the `flask_login.login_required` decorated function.
    """


@overload
def login_required(function: Callable[P, R], /) -> Callable[P, R]:
    """
    Require a user session for the decorated API. No further requirements are applied.
    In effect, this passes along `flask_login.login_required` without modification.

    This can be used as a decorator with `@login_required()`, not `@login_required`, though
    its use case is to wrap a function without using the decorator form, e.g., `wrapped_func = login_required(my_func)`.
    This is the equivalent of `wrapped_func = flask_login.login_required(my_func)`.

    :return Callable[P, R]: Returns the `flask_login.login_required` wrapped function.
    """


@overload
def login_required(
    roles: Sequence[Role | str], /
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Require a user session, and require that the user has at least one of the specified roles.

    :param Sequence[Role  |  str] roles: The list of roles the user can have that will allow them to access the decorated API.
    :return Callable[[Callable[P, R]], Callable[P, R]]: Returns the decorated function.
    """


@overload
def login_required(
    roles: Sequence[Role | str], auth_check_override: AuthCheckOverrideCallable, /
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Require a user session, and require that the user has at least one of the specified roles.

    `auth_check_override` is called to override authorization. If it returns True, the user is considered to have access to the API.
    If it returns False, the roles are checked instead, and the user will have access to the API if they have one of the specified roles.

    :param Sequence[Role  |  str] roles: The list of roles the user can have that will allow them to access the decorated API.
    :param AuthCheckOverrideCallable auth_check_override: The method that is called to override authorization. It receives the following parameters:

        * `user` is the current session user

        * `*args` will be any arguments passed without argument keywords. When using `login_required` as a
          decorator, this will be an empty tuple.

        * `**kwargs` will be any parameters specified with keywords. When using `login_required` as a decorator,
          this will be the parameters passed into the decorated method.
          In the case of a Flask API endpoint, for example, this will be all of the endpoint method parameters.
    :return Callable[[Callable[P, R]], Callable[P, R]]: _description_
    """


def login_required(
    roles: Sequence[Role | str] | Callable[P, R] | Callable[..., Any] | None = None,
    auth_check_override: AuthCheckOverrideCallable | None = None,
    /,
) -> Callable[[Callable[P, R]], Callable[P, R]] | Callable[P, R] | Callable[..., Any]:
    """
    Require a valid Flask session before calling the decorated function.

    This method uses the list of `roles` to determine whether the current session user
    has any of the roles listed. Alternatively, the use of `auth_check_override` is used to
    bypass the role check. If the `auth_check_override` method returns True, the user is considered
    to have access to the decorated API endpoint. If the `auth_check_override` method returns False,
    `login_required` falls back to checking `roles`.

    If `roles` is a list, the user must have been assigned one of the roles specified.
    If `roles` is None, or if the decorator is not explicitly called, this will
    only require a valid Flask session regardless of the user's roles.

    If both `auth_check_override` and `roles` are None, this returns `flask_login.login_required`.
    Call this directly, rather than as a decorator.
    See https://flask-login.readthedocs.io/en/latest/#flask_login.login_required for more information.

    If `roles` is a callable and `auth_check_override` is None, this executes `flask_login.login_required` in its usual way.
    This is the same as using `flask_login.login_required` as a decorator.
    See https://flask-login.readthedocs.io/en/latest/#flask_login.login_required for more information.

    If `roles` is a list of User Roles and the current session user's roles intersect with `roles`,
    this returns True. If the current session user's roles do not intersect with `roles`, this returns False.

    If `auth_check_override` is a callable, it will be called with the following parameters:
        * `user` is the current session user

        * `*args` will be any arguments passed without argument keywords. When using `login_required` as a
          decorator, this will be an empty tuple.

        * `**kwargs` will be any parameters specified with keywords. When using `login_required` as a decorator,
          this will be the parameters passed into the decorated method.
          In the case of a Flask API endpoint, for example, this will be all of the endpoint method parameters.
    """

    if auth_check_override is None:
        if roles is None:
            return flask_login_required

        # In this case, `roles` is actually a function.
        # It is probably a decorated function.
        if callable(roles):
            return flask_login_required(roles)
    else:
        if not callable(auth_check_override):
            raise TypeError("Override must be a callable.")

    if not isinstance(roles, list):
        raise TypeError("Roles must be a list of User Roles.")

    login_manager = FlaskLoginManager()

    def wrapper(fn: Callable[P, R]) -> Callable[P, R | Response]:
        @wraps(fn)
        def decorated_view(*args: P.args, **kwargs: P.kwargs) -> R | Response:
            unauthorized = True
            try:
                user = cast(_LoginUserMixin, current_user)
                if not user.is_authenticated:
                    # this should end up raising a 401 exception
                    return login_manager.unauthorized()

                if callable(auth_check_override):
                    if auth_check_override(user, *args, **kwargs):
                        unauthorized = False

                # if authorization was overriden and successful, don't check roles
                if unauthorized:
                    if not isinstance(roles, list):  # pyright: ignore[reportUnnecessaryIsInstance]
                        # this should end up raising a 401 exception
                        return login_manager.unauthorized()

                    # if roles is empty, no roles will intersect.
                    # this means an empty list means "no roles have access"
                    role_intersection = [
                        str(role)
                        for role in user.roles
                        if (str(role) in ({str(r) for r in roles}) or [])
                    ]
                    if len(role_intersection) == 0:
                        # this should end up raising a 401 exception
                        return login_manager.unauthorized()
            except Exception as e:
                log.exception(e)
                # this should end up raising a 401 exception
                return login_manager.unauthorized()

            return fn(*args, **kwargs)

        return decorated_view

    return wrapper


class Username(TypedDict):
    username: NotRequired[str]


@inject
def get_username(log: Logger) -> Username:
    user = flask_login.current_user
    userId: UserId = user.id
    username = userId.username
    log.debug(f"Flask current user is {username}")
    return {"username": username}


def apikey_auth(token: str, required_scopes: Any):
    """
    This is used by Connexion to authorize requests. It is specified in API-v1.yaml.
    `token` is the value passed either through Connexion ingesting the `session` cookie,
    or through the manually set value in the Swagger UI. The `session` cookie is available
    after authenticating.
    """
    log = logging.getLogger("connexion")

    session_data: Dict[str, Any]
    try:
        session_data = decrypt_flask_cookie(current_app.config["SECRET_KEY"], token)  # pyright: ignore[reportUnknownArgumentType]
        # under normal circumstances these session objects are equivalent.
        # The values will not be equivalent if the cookie is expired and
        # the browser did not send the `session` cookie, but Swagger UI did.
        if session_data["_id"] == session["_id"]:
            log.info("Connexion request has a valid session.")
            return get_username(log)
    except Exception as e:
        log.error(e)

    log.info("Connexion request is unauthorized.")
    raise Unauthorized()


def _delete_username_cookie(response: Response, log: Logger):
    response.delete_cookie(SESSION_VALUE_NAMES.USERNAME)

    # `delete_cookie` doesn't have parameters for "Secure" or "SameSite,"
    # so we add them below.
    headers = response.headers.get("Set-Cookie")
    if headers is None:
        log.error(
            "Headers is empty, but was expected in order to correctly log out a user.\
    The user's browser will retain the `username` cookie."
        )
        return response

    # Using a dict as an ordered set so the cookie contains only unique parts.
    header_parts = dict.fromkeys([
        header_part.strip() for header_part in headers.split(";")
    ])
    header_parts["Secure"] = None
    header_parts["SameSite=None"] = None
    response.headers.set("Set-Cookie", "; ".join(header_parts.keys()))
    return response


sso_blueprint = Blueprint("sso", __name__, url_prefix="/saml")


@dataclass(frozen=True)
class _SessionValueNames:
    USERNAME = "username"
    AUTHENTICATED = "authenticated"
    USER_ID = "user_id"


SESSION_VALUE_NAMES = _SessionValueNames()


# def _SSOBlueprint(self, sso_blueprint: Blueprint):
@sso_blueprint.route("/user")
@flask_login_required
@inject
def user(log: Logger):
    return get_username(log)


@sso_blueprint.route("/logout")
@flask_login_required
@inject
def logout(log: Logger):
    log.info(f"Logging out user {flask_login.current_user}")
    logout_user()
    response = make_response('{"status_code": 200, "status": "200 OK"}', 200)
    response.headers["Content-Type"] = "application/json"

    return _delete_username_cookie(response, log)


@sso_blueprint.route("/login/<idp_name>")
@inject
def sp_initiated(
    idp_name: str, saml2_client: SAML2Client, config: SSOConfig, log: Logger
):
    redirect_url = saml2_client.prepare_user_authentication(
        relay_state=cast(SAML2Config, config.settings).relay_state
    )

    log.info(f"Redirecting client to {redirect_url} for IDP {idp_name}")
    response = redirect(redirect_url, code=302)

    response.headers["Cache-Control"] = "no-cache, no-store"
    response.headers["Pragma"] = "no-cache"
    return response


@sso_blueprint.route("/login/<idp_name>", methods=["POST"])
@inject
def idp_initiated(
    idp_name: str,
    saml2_client: SAML2Client,
    user_loader: UserLoader[UserMixin[Role]],
    config: Config,
    sso_config: SSOConfig,
    log: Logger,
):
    log.info(f"Trying to log in from SAML2 response for IDP {idp_name}")
    saml_response: str = request.form["SAMLResponse"]

    user: UserMixin[Role] | None = None
    try:
        (username, _) = saml2_client.handle_user_login(saml_response)
        log.info(f"Login username is {username}")
        # FIXME should enable a way to set the default role here?
        user = user_loader.user_loader(username, None, True)
        if user is None:
            raise
    except (
        NotValid,
        MustValueError,
        OutsideCardinality,
        ResponseLifetimeExceed,
        ShouldValueError,
        ToEarly,
    ) as e:
        log.exception(e)
        raise BadRequest("SAML2 Response is invalid.")

    session_duration = (
        config.flask.session.lifetime if config.flask and config.flask.session else None
    )

    login_user(user, remember=False, duration=session_duration)

    from flask import session

    session[SESSION_VALUE_NAMES.AUTHENTICATED] = True
    session[SESSION_VALUE_NAMES.USERNAME] = user.id.username
    session[SESSION_VALUE_NAMES.USER_ID] = user.id.user_id
    log.info(f"Login for user {user.id.user_id} succeeded")

    url: str = ""
    if "RelayState" in request.form:
        redirect_url: str = request.form["RelayState"]
        expected_url = cast(SAML2Config, sso_config.settings).relay_state
        parsed_redirect_url = urlparse(redirect_url)
        parsed_expected_url = urlparse(expected_url)
        if (
            parsed_redirect_url.scheme != parsed_expected_url.scheme
            or parsed_redirect_url.netloc != parsed_expected_url.netloc
        ):
            log.error(
                f'Parsed URL "{redirect_url}" does not match expected URL "{expected_url}"'
            )
            raise Forbidden()
        url: str = request.form["RelayState"]
        log.info(f'SAML2 response has RelayState="{url}"')

    if url == "":
        url = url_for("sso.user")
        log.info(
            f"SAML2 response does not have RelayState. Redirecting client to {url}"
        )

    response = redirect(url)

    response.set_cookie(
        "username",
        str(user.id.username),
        samesite="None",
        secure=True,
        httponly=False,
        max_age=session_duration,
    )
    return response


# @sso_blueprint.before_app_request  # pyright: ignore[reportUntypedFunctionDecorator]
@inject
def make_session_permanent(
    config: Config,
):
    from flask import session

    session.permanent = (
        config.flask.session.permanent
        if config.flask and config.flask.session
        else False
    )


# FIXME these after_app_requests need to be made middleware
# @sso_blueprint.after_app_request  # pyright: ignore[reportUntypedFunctionDecorator]
@inject
def remove_username_cookie_without_session(response: Response, log: Logger):
    from flask import session

    # if the session has expired but the client still has the username cookie,
    # the username cookie needs to be cleared from the client
    if not session.get(  # pyright: ignore[reportUnknownMemberType]
        SESSION_VALUE_NAMES.AUTHENTICATED
    ) and request.cookies.get(SESSION_VALUE_NAMES.USERNAME):
        log.info("Session expired; clearing username cookie.")
        return _delete_username_cookie(response, log)

    return response


# TODO this should probably be moved into the tests
class LoginManager(FlaskLoginManager):
    @inject
    def __init__(
        self,
        user_loader: UserLoader[UserMixin[Role]],
        app: Flask,
        log: Logger,
        add_context_processor: bool = True,
    ):
        super().__init__(app=app, add_context_processor=add_context_processor)

        self._app = app
        self._log = log

        user_loader_lambda: Callable[[str], UserMixin[Role] | None] = (
            lambda username: user_loader.user_loader(username, None)
        )
        _ = self.user_loader(user_loader_lambda)
        self.init_app(app)

    @override
    def unauthorized(self):
        """
        Raises an `Unauthorized` exception.
        """
        saml_idp_url = url_for("sso.sp_initiated", idp_name="okta", _external=True)
        self._log.debug(
            f"Redirecting unauthenticated client to {saml_idp_url} from {request.url}"
        )
        response: Response = self._app.make_response((
            "Unauthorized",
            401,
            {
                "Content-Type": "text/plain; charset=utf-8",
                # Clients can access this IDP URL to authenticate and POST back with a valid SAML request.
                "Location": saml_idp_url,
                # "SAML" isn't in the HTTP standard, but it's the only auth method supported by CAP.
                "WWW-Authenticate": 'SAML realm="Authenticate with the SAML IDP at the URL provided by Location.", charset="UTF-8"',
                "Access-Control-Expose-Headers": "Location, WWW-Authenticate",
            },
        ))
        raise Unauthorized(response.data, response)


class SAML2MiddlewareModule(ConfigurableModule):  # Module):
    @override
    @staticmethod
    def get_config_type() -> type[AbstractConfig]:
        return SSOConfig

    @override
    def configure(self, binder: Binder) -> None:
        binder.install(SAML2Module)

    def register_middleware(self, app: FlaskApp):
        app.add_middleware(SAML2MiddlewareModule.SAML2Middleware)

    class SAML2Middleware:
        _app: ASGIApp

        def __init__(self, app: ASGIApp):
            super().__init__()
            self._app = app

        @inject
        async def __call__(
            self,
            scope: Scope,
            receive: Receive,
            send: Send,
            app: Flask,
            injector: Injector,
            log: Logger,
        ) -> None:
            async def wrapped_send(message: Any) -> None:
                # Only run during startup of the application
                if (
                    scope["type"] != "lifespan"
                    or message["type"] != "lifespan.startup.complete"
                    or not scope["app"]
                ):
                    return await send(message)

                injector.binder.bind(SSOModule, to=SAML2Module())

                log.debug("Registering SSO blueprint.")
                app.register_blueprint(sso_blueprint)
                # trigger side-effect - adds login_manager to app
                _ = injector.get(LoginManager)
                log.debug("SSO blueprint registered and LoginManager installed.")
                return await send(message)

            await self._app(scope, receive, wrapped_send)



"""
This type stub file was generated by pyright.
"""

from jsonschema import Draft4Validator

logger = ...
TYPE_MAP = ...
draft4_format_checker = Draft4Validator.FORMAT_CHECKER
class ParameterValidator:
    def __init__(self, parameters, uri_parser, strict_validation=..., security_query_params=...) -> None:
        """
        :param parameters: List of request parameter dictionaries
        :param uri_parser: class to use for uri parsing
        :param strict_validation: Flag indicating if parameters not in spec are allowed
        :param security_query_params: List of query parameter names used for security.
            These parameters will be ignored when checking for extra parameters in case of
            strict validation.
        """
        ...
    
    @staticmethod
    def validate_parameter(parameter_type, value, param, param_name=...): # -> str | None:
        ...
    
    @staticmethod
    def validate_parameter_list(request_params, spec_params): # -> set[Any]:
        ...
    
    def validate_query_parameter_list(self, request, security_params=...): # -> set[Any]:
        ...
    
    def validate_query_parameter(self, param, request): # -> str | None:
        """
        Validate a single query parameter (request.args in Flask)

        :type param: dict
        :rtype: str
        """
        ...
    
    def validate_path_parameter(self, param, request): # -> str | None:
        ...
    
    def validate_header_parameter(self, param, request): # -> str | None:
        ...
    
    def validate_cookie_parameter(self, param, request): # -> str | None:
        ...
    
    def validate(self, scope): # -> None:
        ...
    
    def validate_request(self, request): # -> None:
        ...
    





"""
This type stub file was generated by pyright.
"""

from .main import ASGIDecorator, FlaskDecorator, StarletteDecorator, WSGIDecorator

"""
This module defines decorators which Connexion uses to wrap user provided view functions.
"""

__all__ = ("ASGIDecorator", "FlaskDecorator", "StarletteDecorator", "WSGIDecorator")



"""
This type stub file was generated by pyright.
"""

from starlette.types import Receive, Scope
from connexion.operations import AbstractOperation

class TestContext:
    __test__ = ...
    def __init__(self, *, context: dict = ..., operation: AbstractOperation = ..., receive: Receive = ..., scope: Scope = ...) -> None:
        ...
    
    def __enter__(self) -> None:
        ...
    
    def __exit__(self, type, value, traceback): # -> Literal[False]:
        ...
    
    @staticmethod
    def build_context() -> dict:
        ...
    
    @staticmethod
    def build_operation() -> AbstractOperation:
        ...
    
    @staticmethod
    def build_receive() -> Receive:
        ...
    
    @staticmethod
    def build_scope(**kwargs) -> Scope:
        ...
    





import logging
import re
from dataclasses import asdict
from typing import Any, cast

import pytest
from Ligare.web.scaffolding.__main__ import ScaffolderCli, scaffold
from Ligare.web.scaffolding.scaffolder import (
    Operation,
    ScaffoldConfig,
    ScaffoldEndpoint,
    Scaffolder,
    ScaffoldModule,
)
from pytest import CaptureFixture
from pytest_mock import MockerFixture

# pyright: reportPrivateUsage=false


@pytest.mark.parametrize(
    "name,expected_name",
    [
        ("foo", "foo"),
        ("FOO", "foo"),
        ("foo_bar", "foo_bar"),
        ("foo-bar", "foo_bar"),
        ("foo.bar", "foo_bar"),
        ("foo bar", "foo_bar"),
        ("foo0bar", "foo0bar"),
    ],
)
def test__operation__normalizes_module_name(name: str, expected_name: str):
    operation = Operation(name)
    assert expected_name == operation.module_name


@pytest.mark.parametrize(
    "name,expected_name",
    [
        ("foo", "foo"),
        ("FOO", "foo"),
        ("foo_bar", "foo_bar"),
        ("foo-bar", "foo-bar"),
        ("foo.bar", "foo.bar"),
        ("foo bar", "foo bar"),
        ("foo0bar", "foo0bar"),
    ],
)
def test__operation__normalizes_url_path_name(name: str, expected_name: str):
    operation = Operation(name)
    assert expected_name == operation.url_path_name


@pytest.mark.parametrize(
    "name",
    ["foo", "FOO", "foo_bar", "foo-bar", "foo.bar", "foo bar", "foo0bar"],
)
def test__operation__does_not_normalize_raw_name(name: str):
    operation = Operation(name)
    assert name == operation.raw_name


def test__parse_args__requires_run_mode_switch(capsys: CaptureFixture[str]):
    with pytest.raises(SystemExit) as e:
        _ = ScaffolderCli()._parse_args([])
    captured = capsys.readouterr()
    assert e.value.code == 2
    assert (
        "error: the following arguments are required: {create,modify}" in captured.err
    )


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__modes_requires_name(mode: str, capsys: CaptureFixture[str]):
    with pytest.raises(SystemExit) as e:
        _ = ScaffolderCli()._parse_args([mode])
    captured = capsys.readouterr()
    assert e.value.code == 2
    assert "error: the following arguments are required: -n" in captured.err


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__supports_specific_run_modes(mode: str):
    args = ScaffolderCli()._parse_args([mode, "-n", "test"])
    assert mode == args.mode


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__disallows_application_to_be_named_application(
    mode: str, capsys: CaptureFixture[str]
):
    with pytest.raises(SystemExit) as e:
        _ = ScaffolderCli()._parse_args([mode, "-n", "application"])

    captured = capsys.readouterr()
    assert e.value.code == 2
    assert (
        "`application` is not an allowed value of the `name` argument." in captured.err
    )


@pytest.mark.parametrize(
    "mode,application_name,expected_url_path_name,expected_module_name",
    [
        ("create", "FOO", "foo", "foo"),
        ("create", "foo", "foo", "foo"),
        ("create", "FOO_BAR", "foo_bar", "foo_bar"),
        ("create", "FOO-BAR", "foo-bar", "foo_bar"),
        ("create", "FOO BAR", "foo bar", "foo_bar"),
        ("modify", "FOO", "foo", "foo"),
        ("modify", "foo", "foo", "foo"),
        ("modify", "FOO_BAR", "foo_bar", "foo_bar"),
        ("modify", "FOO-BAR", "foo-bar", "foo_bar"),
        ("modify", "FOO BAR", "foo bar", "foo_bar"),
    ],
)
def test__parse_args__normalizes_application_name(
    mode: str,
    application_name: str,
    expected_url_path_name: str,
    expected_module_name: str,
):
    args = ScaffolderCli()._parse_args([mode, "-n", application_name])

    assert args.name is not None
    assert expected_url_path_name == args.name.url_path_name
    assert expected_module_name == args.name.module_name


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__disallows_endpoints_named_application(
    mode: str, capsys: CaptureFixture[str]
):
    with pytest.raises(SystemExit) as e:
        _ = ScaffolderCli()._parse_args([mode, "-n", "test", "-e", "application"])

    captured = capsys.readouterr()
    assert e.value.code == 2
    assert (
        "`application` is not an allowed value of the `endpoint` argument."
        in captured.err
    )


@pytest.mark.parametrize(
    "mode,endpoint_name,expected_url_path_name,expected_module_name",
    [
        ("create", "FOO", "foo", "foo"),
        ("create", "foo", "foo", "foo"),
        ("create", "FOO_BAR", "foo_bar", "foo_bar"),
        ("create", "FOO-BAR", "foo-bar", "foo_bar"),
        ("create", "FOO BAR", "foo bar", "foo_bar"),
        ("modify", "FOO", "foo", "foo"),
        ("modify", "foo", "foo", "foo"),
        ("modify", "FOO_BAR", "foo_bar", "foo_bar"),
        ("modify", "FOO-BAR", "foo-bar", "foo_bar"),
        ("modify", "FOO BAR", "foo bar", "foo_bar"),
    ],
)
def test__parse_args__normalizes_endpoint_names(
    mode: str,
    endpoint_name: str,
    expected_url_path_name: str,
    expected_module_name: str,
):
    args = ScaffolderCli()._parse_args([mode, "-n", "test", "-e", endpoint_name])

    assert args.endpoints is not None
    assert (expected_url_path_name, expected_module_name) in {
        (endpoint.url_path_name, endpoint.module_name) for endpoint in args.endpoints
    }


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__allows_multiple_endpoints(mode: str):
    args = ScaffolderCli()._parse_args([mode, "-n", "test", "-e", "foo", "-e", "bar"])

    assert args.endpoints is not None
    endpoints = {
        (endpoint.url_path_name, endpoint.module_name) for endpoint in args.endpoints
    }
    assert ("foo", "foo") in endpoints
    assert ("bar", "bar") in endpoints


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__disallows_duplicated_endpoint_names(
    mode: str, capsys: CaptureFixture[str]
):
    with pytest.raises(SystemExit) as e:
        # foo-bar and foo_bar are normalized and are equivalent.
        # no need to duplicate the name normalization tests; just use the exact values here
        # fmt: off
        _ = ScaffolderCli()._parse_args([mode, "-n", "test", "-e", "foo", "-e", "foo_bar", "-e", "foo-bar"])
        # fmt: on
    captured = capsys.readouterr()
    assert e.value.code == 2
    assert (
        f"{mode}: error: argument -e: The ['-e'] argument does not allow duplicate values. The value `foo-bar` duplicates the value `foo_bar`."
        in captured.err
    )


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__allows_endpoints_with_same_name_as_application(mode: str):
    # foo-bar and foo_bar are normalized and are equivalent.
    # no need to duplicate the name normalization tests; just use the exact values here
    _ = ScaffolderCli()._parse_args([mode, "-n", "foo-bar", "-e", "foo_bar"])


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__uses_application_name_for_endpoint_if_no_endpoints_given(
    mode: str,
):
    args = ScaffolderCli()._parse_args([mode, "-n", "test"])

    assert args.endpoints is not None
    assert ("test", "test") in {
        (endpoint.url_path_name, endpoint.module_name) for endpoint in args.endpoints
    }


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__does_not_use_application_name_for_endpoint_if_endpoints_given(
    mode: str,
):
    args = ScaffolderCli()._parse_args([mode, "-n", "test", "-e", "foo"])

    assert args.endpoints is not None
    endpoints = {
        (endpoint.url_path_name, endpoint.module_name) for endpoint in args.endpoints
    }

    assert ("test", "test") not in endpoints
    assert ("foo", "foo") in endpoints


@pytest.mark.parametrize(
    "mode,template_type",
    [
        ("create", "basic"),
        ("create", "openapi"),
    ],
)
def test__parse_args__supports_specific_template_types(
    mode: str,
    template_type: str,
):
    args = ScaffolderCli()._parse_args([mode, "-n", "test", "-t", template_type])

    assert template_type == args.template_type


@pytest.mark.parametrize(
    "mode,template_type",
    [
        ("create", "BASIC"),
        ("create", "basic"),
        ("create", "OPENAPI"),
        ("create", "openapi"),
    ],
)
def test__parse_args__lowercases_template_type(
    mode: str,
    template_type: str,
):
    args = ScaffolderCli()._parse_args([mode, "-n", "test", "-t", template_type])

    assert template_type.lower() == args.template_type


@pytest.mark.parametrize("mode,module_name", [("create", "database")])
def test__parse_args__supports_specific_modules(
    mode: str,
    module_name: str,
):
    args = ScaffolderCli()._parse_args([mode, "-n", "test", "-m", module_name])

    assert args.modules is not None
    assert module_name in args.modules


@pytest.mark.parametrize(
    "mode,module_name", [("create", "DATABASE"), ("create", "TEST")]
)
def test__parse_args__does_not_lowercase_modules(
    mode: str, module_name: str, capsys: CaptureFixture[str]
):
    with pytest.raises(SystemExit) as e:
        _ = ScaffolderCli()._parse_args([mode, "-n", "test", "-m", module_name])

    captured = capsys.readouterr()
    assert e.value.code == 2
    assert (
        re.search(
            rf"{mode}: error: argument -m: invalid choice: '{module_name}' \(choose from.+'{module_name.lower()}'",
            captured.err,
        )
        is not None
    )


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__supports_setting_explicit_output_directory(mode: str):
    args = ScaffolderCli()._parse_args([mode, "-n", "test", "-o", "output_dir"])

    assert "output_dir" == args.output_directory


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__parse_args__uses_application_name_for_output_directory_if_no_outout_directory_given(
    mode: str,
):
    args = ScaffolderCli()._parse_args([mode, "-n", "test"])

    assert "test" == args.output_directory


@pytest.mark.parametrize("env_log_level", [str(logging.DEBUG), str(logging.INFO)])
def test__scaffold__sets_log_level_from_envvar_when_parameter_not_specified(
    env_log_level: str,
    mocker: MockerFixture,
):
    scaffolder_cli_mock = mocker.patch("Ligare.web.scaffolding.__main__.ScaffolderCli")
    environ_mock = mocker.patch(
        "Ligare.web.scaffolding.__main__.environ.get",
        return_value=env_log_level,
    )

    scaffold()

    environ_mock.assert_called_once_with("LOG_LEVEL")
    scaffolder_cli_mock.assert_called_with(env_log_level)


@pytest.mark.parametrize("log_level", ["DEBUG", "debug", "10", 10])
def test__scaffold__sets_log_level_from_value_when_parameter_specified(
    log_level: str | int, mocker: MockerFixture
):
    _ = mocker.patch("Ligare.web.scaffolding.__main__.ScaffolderCli.run")
    _ = mocker.patch("Ligare.web.scaffolding.__main__.environ")

    log_mock = mocker.patch("Ligare.web.scaffolding.__main__.logging.basicConfig")

    expected_log_level = 10
    scaffold(log_level=log_level)

    log_mock.assert_called()
    for call_args in log_mock.call_args:
        if not isinstance(call_args, dict):
            continue
        assert "level" in call_args
        assert expected_log_level == call_args["level"]


def test__scaffold__fails_when_log_level_is_any_unexpected_exception(
    mocker: MockerFixture,
):
    _ = mocker.patch("Ligare.web.scaffolding.__main__.ScaffolderCli.run")

    def fake_exception(*args: Any, **kwargs: Any):
        raise ValueError("fake exception")

    _ = mocker.patch("Ligare.web.scaffolding.__main__.int", side_effect=fake_exception)

    with pytest.raises(ValueError, match=r"^fake exception$"):
        scaffold(log_level="x")


def test__scaffold__uses_sys_argv_when_args_not_given(mocker: MockerFixture):
    scaffolder_cli = mocker.patch("Ligare.web.scaffolding.__main__.ScaffolderCli.run")
    argv_values = ["1", "2", "3"]
    argv = mocker.patch("Ligare.web.scaffolding.__main__.sys.argv")
    argv.__getitem__ = (
        lambda _argv, _slice: argv_values  # pyright: ignore[reportUnknownLambdaType]
    )

    scaffold()

    scaffolder_cli.assert_called_with(argv_values)


def test__scaffold__uses_argv_when_sys_argv_not_set(mocker: MockerFixture):
    scaffolder_cli = mocker.patch("Ligare.web.scaffolding.__main__.ScaffolderCli.run")
    argv_values = ["1", "2", "3"]

    scaffold(argv_values)

    scaffolder_cli.assert_called_with(argv_values)


@pytest.mark.parametrize(
    "mode,config_name,config_value",
    [
        ("create", "mode", "create"),
        ("create", "template_type", "basic"),
        ("create", "output_directory", "test"),
        ("create", "application.module_name", "test"),
        ("modify", "mode", "modify"),
        ("modify", "output_directory", "test"),
        ("modify", "application.module_name", "test"),
    ],
)
def test__scaffold__uses_argv_for_basic_scaffold_configuration(
    mode: str, config_name: str, config_value: str, mocker: MockerFixture
):
    _ = mocker.patch("Ligare.web.scaffolding.__main__.Scaffolder")
    config_mock = mocker.patch("Ligare.web.scaffolding.__main__.ScaffoldConfig")

    argv_values = [mode, "-n", "test"]

    scaffold(argv_values)

    config = config_mock.call_args.kwargs

    config_name_parts = config_name.split(".")
    if len(config_name_parts) == 1:
        assert config_value == config[config_name_parts[0]]
    else:
        # tests `config['application']['module_name']`
        assert (
            config_value == asdict(config[config_name_parts[0]])[config_name_parts[1]]
        )


@pytest.mark.parametrize("mode", ["create", "modify"])
def test__scaffold__uses_argv_for_endpoint_configuration(
    mode: str,
    mocker: MockerFixture,
):
    _ = mocker.patch("Ligare.web.scaffolding.__main__.Scaffolder")
    config_mock = mocker.patch("Ligare.web.scaffolding.__main__.ScaffoldConfig")

    argv_values = [mode, "-n", "test", "-e", "foo", "-e", "bar"]

    scaffold(argv_values)

    endpoint_names = [
        endpoint.operation.url_path_name
        for endpoint in cast(
            list[ScaffoldEndpoint], config_mock.call_args.kwargs["endpoints"]
        )
    ]

    assert "foo" in endpoint_names
    assert "bar" in endpoint_names


def test__scaffold__create_mode_uses_argv_for_module_configuration(
    mocker: MockerFixture,
):
    _ = mocker.patch("Ligare.web.scaffolding.__main__.Scaffolder")
    config_mock = mocker.patch("Ligare.web.scaffolding.__main__.ScaffoldConfig")

    argv_values = ["create", "-n", "test", "-m", "database"]

    scaffold(argv_values)

    module_names = [
        module.module_name
        for module in cast(
            list[ScaffoldModule], config_mock.call_args.kwargs["modules"]
        )
    ]

    assert "database" in module_names


def test__scaffold__create_mode_database_module_configures_database(
    mocker: MockerFixture,
):
    # TODO refactor Scaffolder so all these mocks don't need to happen
    _ = mocker.patch("builtins.input", return_value="Y")
    _ = mocker.patch("Ligare.web.scaffolding.scaffolder.BaseLoader")
    _ = mocker.patch("Ligare.web.scaffolding.scaffolder.Environment")
    _ = mocker.patch("Ligare.web.scaffolding.scaffolder.PackageLoader")
    _ = mocker.patch("Ligare.web.scaffolding.scaffolder.Template")
    _ = mocker.patch("Ligare.web.scaffolding.scaffolder.Scaffolder._create_directory")
    _ = mocker.patch("Ligare.web.scaffolding.scaffolder.Scaffolder._scaffold_directory")
    _ = mocker.patch("Ligare.web.scaffolding.scaffolder.Scaffolder._scaffold_endpoints")
    _ = mocker.patch(
        "Ligare.web.scaffolding.scaffolder.Scaffolder._check_scaffolded_application_exists",
        return_value=False,
    )
    module_hook_mock = mocker.patch(
        "Ligare.web.scaffolding.scaffolder.Scaffolder._execute_module_hooks"
    )

    argv_values = ["create", "-n", "test", "-m", "database", "-o", "/dev/null"]

    scaffold(argv_values)

    pass
    module_hook_mock.assert_called_once_with(
        "scaffolding/templates/optional/{{application.module_name}}/modules/database"
    )


@pytest.fixture
def scaffold_config():
    return ScaffoldConfig(
        output_directory="/dev/null",
        application=Operation(name="foo"),
        template_type="basic",
        modules=[ScaffoldModule(module_name="database")],
    )


def test__database_module__on_create__asks_for_connection_string(
    scaffold_config: ScaffoldConfig,
    mocker: MockerFixture,
):
    input_mock = mocker.patch("builtins.input", return_value="")

    mock_log = mocker.patch("logging.Logger", spec=logging.Logger)
    scaffolder = Scaffolder(config=scaffold_config, log=mock_log)
    scaffolder._execute_module_hooks(
        "scaffolding/templates/optional/{{application.module_name}}/modules/database",
    )

    input_mock.assert_called()
    assert [
        arg
        for arg in input_mock.call_args.args
        if "Enter a database connection string" in arg
    ]


def test__database_module__on_create__uses_user_input_connection_string(
    scaffold_config: ScaffoldConfig,
    mocker: MockerFixture,
):
    _ = mocker.patch("builtins.input", return_value="sqlite:///dev/null")

    mock_log = mocker.patch("logging.Logger", spec=logging.Logger)
    scaffolder = Scaffolder(config=scaffold_config, log=mock_log)
    scaffolder._execute_module_hooks(
        "scaffolding/templates/optional/{{application.module_name}}/modules/database",
    )

    assert (
        scaffolder._config_dict["module"]["database"]["connection_string"]
        == "sqlite:///dev/null"
    )



"""
This type stub file was generated by pyright.
"""

import typing as t
import flask
import werkzeug.routing
from connexion import jsonifier
from connexion.frameworks.abstract import Framework
from connexion.lifecycle import WSGIRequest
from connexion.uri_parsing import AbstractURIParser

"""
This module defines functionality specific to the Flask framework.
"""
class Flask(Framework):
    @staticmethod
    def is_framework_response(response: t.Any) -> bool:
        ...
    
    @classmethod
    def connexion_to_framework_response(cls, response): # -> flask.wrappers.Response | werkzeug.sansio.response.Response:
        ...
    
    @classmethod
    def build_response(cls, data: t.Any, *, content_type: str = ..., headers: dict = ..., status_code: int = ...): # -> flask.wrappers.Response | werkzeug.sansio.response.Response:
        ...
    
    @staticmethod
    def get_request(*, uri_parser: AbstractURIParser, **kwargs) -> WSGIRequest:
        ...
    


PATH_PARAMETER = ...
PATH_PARAMETER_CONVERTERS = ...
def flaskify_endpoint(identifier, randomize=...): # -> str:
    """
    Converts the provided identifier in a valid flask endpoint name

    :type identifier: str
    :param randomize: If specified, add this many random characters (upper case
        and digits) to the endpoint name, separated by a pipe character.
    :type randomize: int | None
    :rtype: str

    """
    ...

def convert_path_parameter(match, types): # -> str:
    ...

def flaskify_path(swagger_path, types=...): # -> str:
    """
    Convert swagger path templates to flask path templates

    :type swagger_path: str
    :type types: dict
    :rtype: str

    >>> flaskify_path('/foo-bar/{my-param}')
    '/foo-bar/<my_param>'

    >>> flaskify_path('/foo/{someint}', {'someint': 'int'})
    '/foo/<int:someint>'
    """
    ...

class FlaskJSONProvider(flask.json.provider.DefaultJSONProvider):
    """Custom JSONProvider which adds connexion defaults on top of Flask's"""
    @jsonifier.wrap_default
    def default(self, o): # -> Any:
        ...
    


class NumberConverter(werkzeug.routing.BaseConverter):
    """Flask converter for OpenAPI number type"""
    regex = ...
    def to_python(self, value): # -> float:
        ...
    


class IntegerConverter(werkzeug.routing.BaseConverter):
    """Flask converter for OpenAPI integer type"""
    regex = ...
    def to_python(self, value): # -> int:
        ...
    





"""
This type stub file was generated by pyright.
"""

class UserMixin:
    """
    This provides default implementations for the methods that Flask-Login
    expects user objects to have.
    """
    __hash__ = ...
    @property
    def is_active(self) -> bool:
        ...
    
    @property
    def is_authenticated(self) -> bool:
        ...
    
    @property
    def is_anonymous(self) -> bool:
        ...
    
    def get_id(self) -> str:
        ...
    
    def __eq__(self, other: object) -> bool:
        """
        Checks the equality of two `UserMixin` objects using `get_id`.
        """
        ...
    
    def __ne__(self, other: object) -> bool:
        """
        Checks the inequality of two `UserMixin` objects using `get_id`.
        """
        ...
    


class AnonymousUserMixin:
    """
    This is the default object for representing an anonymous user.
    """
    @property
    def is_authenticated(self) -> bool:
        ...
    
    @property
    def is_active(self) -> bool:
        ...
    
    @property
    def is_anonymous(self) -> bool:
        ...
    
    def get_id(self) -> None:
        ...
    





from typing import Any, Callable, Generator

import pytest
from Ligare.database.testing.config import (
    UseInmemoryDatabaseLoader,
    inmemory_database_config_loader,
)
from Ligare.programming.collections.dict import AnyDict
from Ligare.programming.config import TConfig
from pytest_mock import MockerFixture

TConfigCallable = Callable[[type[TConfig], str, AnyDict | None], TConfig]
UseInmemoryDatabaseResult = Generator[
    tuple[UseInmemoryDatabaseLoader, UseInmemoryDatabaseLoader], Any, None
]


@pytest.fixture
def use_inmemory_database(
    mocker: MockerFixture,
) -> UseInmemoryDatabaseResult:
    loader = inmemory_database_config_loader()
    yield (
        mocker.patch("Ligare.web.application.load_config", loader),
        mocker.patch("Ligare.database.migrations.alembic.env.load_config", loader),
    )
